"""Streaming utilities."""
from .cache import Cache
from .iter_arff import iter_arff
from .iter_array import iter_array
from .iter_csv import iter_csv
from .iter_libsvm import iter_libsvm
from .qa import simulate_qa
from .shuffling import shuffle

__all__ = [
    'Cache',
    'iter_arff',
    'iter_array',
    'iter_csv',
    'iter_libsvm',
    'simulate_qa',
    'shuffle'
]

try:
    from .iter_pandas import iter_pandas
    __all__ += ['iter_pandas']
except ImportError:
    pass

try:
    from .iter_sklearn import iter_sklearn_dataset
    __all__ += ['iter_sklearn_dataset']
except ImportError:
    pass

try:
    from .iter_sql import iter_sql
    __all__ += ['iter_sql']
except ImportError:
    pass

try:
    from .iter_vaex import iter_vaex
    __all__ += ['iter_vaex']
except ImportError:
    pass



import csv
import datetime as dt
import os
import random
import typing

from .. import base

from . import utils


__all__ = ['iter_csv']


class DictReader(csv.DictReader):
    """Overlay on top of `csv.DictReader` which allows sampling."""

    def __init__(self, fraction, rng, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fraction = fraction
        self.rng = rng

    def __next__(self):

        if self.line_num == 0:
            self.fieldnames

        row = next(self.reader)

        if self.fraction < 1:
            while self.rng.random() > self.fraction:
                row = next(self.reader)

        return dict(zip(self.fieldnames, row))


def iter_csv(filepath_or_buffer, target: typing.Union[str, typing.List[str]] = None,
             converters: dict = None, parse_dates: dict = None, drop: typing.List[str] = None,
             fraction=1., compression='infer', seed: int = None, field_size_limit: int = None,
             **kwargs) -> base.typing.Stream:
    """Iterates over rows from a CSV file.

    Parameters:
        filepath_or_buffer: Either a string indicating the location of a CSV file, or a buffer
            object that has a `read` method.
        target: A single target column is assumed if a string is passed. A multiple output scenario
            is assumed if a list of strings is passed. A `None` value will be assigned to each `y`
            if this parameter is omitted.
        converters: A `dict` mapping feature names to callables used to parse their associated
            values.
        parse_dates: A `dict` mapping feature names to a format passed to the
            `datetime.datetime.strptime` method.
        drop: Fields to ignore.
        fraction: Sampling fraction.
        compression: For on-the-fly decompression of on-disk data. If this is set to 'infer' and
            `filepath_or_buffer` is a path, then the decompression method is inferred for the
            following extensions: '.gz', '.zip'.
        seed: If specified, the sampling will be deterministic.
        field_size_limit: If not `None`, this will be passed to the `csv.field_size_limit`
            function.
        kwargs: All other keyword arguments are passed to the underlying `csv.DictReader`.

    Returns:
        By default each feature value will be of type `str`. You can use the `converters` and
        `parse_dates` parameters to convert them as you see fit.

    Example:

        Although this function is designed to handle different kinds of inputs, the most common
        use case is to read a file on the disk. We'll first create a little CSV file to illustrate.

        >>> tv_shows = '''name,year,rating
        ... Planet Earth II,2016,9.5
        ... Planet Earth,2006,9.4
        ... Band of Brothers,2001,9.4
        ... Breaking Bad,2008,9.4
        ... Chernobyl,2019,9.4
        ... '''
        >>> with open('tv_shows.csv', mode='w') as f:
        ...     _ = f.write(tv_shows)

        We can now go through the rows one by one. We can use the `converters` parameter to cast
        the `rating` field value as a `float`. We can also convert the `year` to a `datetime` via
        the `parse_dates` parameter.

        >>> from creme import stream

        >>> params = {
        ...     'converters': {'rating': float},
        ...     'parse_dates': {'year': '%Y'}
        ... }
        >>> for x, y in stream.iter_csv('tv_shows.csv', **params):
        ...     print(x, y)
        {'name': 'Planet Earth II', 'year': datetime.datetime(2016, 1, 1, 0, 0), 'rating': 9.5} None
        {'name': 'Planet Earth', 'year': datetime.datetime(2006, 1, 1, 0, 0), 'rating': 9.4} None
        {'name': 'Band of Brothers', 'year': datetime.datetime(2001, 1, 1, 0, 0), 'rating': 9.4} None
        {'name': 'Breaking Bad', 'year': datetime.datetime(2008, 1, 1, 0, 0), 'rating': 9.4} None
        {'name': 'Chernobyl', 'year': datetime.datetime(2019, 1, 1, 0, 0), 'rating': 9.4} None

        The value of `y` is always `None` because we haven't provided a value for the `target`
        parameter. Here is an example where a `target` is provided:

        >>> X_y = stream.iter_csv('tv_shows.csv', target='rating', **params)
        >>> for x, y in X_y:
        ...     print(x, y)
        {'name': 'Planet Earth II', 'year': datetime.datetime(2016, 1, 1, 0, 0)} 9.5
        {'name': 'Planet Earth', 'year': datetime.datetime(2006, 1, 1, 0, 0)} 9.4
        {'name': 'Band of Brothers', 'year': datetime.datetime(2001, 1, 1, 0, 0)} 9.4
        {'name': 'Breaking Bad', 'year': datetime.datetime(2008, 1, 1, 0, 0)} 9.4
        {'name': 'Chernobyl', 'year': datetime.datetime(2019, 1, 1, 0, 0)} 9.4

        Finally, let's delete the example file.

        >>> import os; os.remove('tv_shows.csv')

    .. tip::
        Reading CSV files can be quite slow. If, for whatever reason, you're going to loop through
        the same file multiple times, then we recommend that you to use the `stream.Cache` utility.

    """

    # Set the field size limit
    limit = csv.field_size_limit()
    if field_size_limit is not None:
        csv.field_size_limit(field_size_limit)

    # If a file is not opened, then we open it
    buffer = filepath_or_buffer
    if not hasattr(buffer, 'read'):
        buffer = utils.open_filepath(buffer, compression)

    for x in DictReader(fraction=fraction, rng=random.Random(seed), f=buffer, **kwargs):

        if drop:
            for i in drop:
                del x[i]

        # Cast the values to the given types
        if converters is not None:
            for i, t in converters.items():
                x[i] = t(x[i])

        # Parse the dates
        if parse_dates is not None:
            for i, fmt in parse_dates.items():
                x[i] = dt.datetime.strptime(x[i], fmt)

        # Separate the target from the features
        y = None
        if isinstance(target, list):
            y = {name: x.pop(name) for name in target}
        elif target is not None:
            y = x.pop(target)

        yield x, y

    # Close the file if we opened it
    if buffer is not filepath_or_buffer:
        buffer.close()

    # Reset the file size limit to it's original value
    csv.field_size_limit(limit)



import collections
import types

try:
    import graphviz
    GRAPHVIZ_INSTALLED = True
except ImportError:
    GRAPHVIZ_INSTALLED = False

from creme import base

from . import func


__all__ = ['TransformerUnion']


class TransformerUnion(base.Transformer):
    """Packs multiple transformers into a single one.

    Pipelines allow you to apply steps sequentially. Therefore, the output of a step becomes the
    input of the next one. In many cases, you may want to pass the output of a step to multiple
    steps. This simple transformer allows you to do so. In other words, it enables you to apply
    particular steps to different parts of an input. A typical example is when you want to scale
    numeric features and one-hot encode categorical features.

    This transformer is essentially a list of transformers. Whenever it is updated, it loops
    through each transformer and updates them. Meanwhile, calling `transform_one` collects the
    output of each transformer and merges them into a single dictionary.

    Parameters:
        transformers: Ideally, a list of (name, estimator) tuples. A name is automatically inferred
            if none is provided.

    Example:

        Take the following dataset:

        >>> X = [
        ...     {'place': 'Taco Bell', 'revenue': 42},
        ...     {'place': 'Burger King', 'revenue': 16},
        ...     {'place': 'Burger King', 'revenue': 24},
        ...     {'place': 'Taco Bell', 'revenue': 58},
        ...     {'place': 'Burger King', 'revenue': 20},
        ...     {'place': 'Taco Bell', 'revenue': 50}
        ... ]

        As an example, let's assume we want to compute two aggregates of a dataset. We therefore
        define two `feature_extraction.Agg`s and initialize a `TransformerUnion` with them:

        >>> from creme import compose
        >>> from creme import feature_extraction
        >>> from creme import stats

        >>> mean = feature_extraction.Agg(
        ...     on='revenue', by='place',
        ...     how=stats.Mean()
        ... )
        >>> count = feature_extraction.Agg(
        ...     on='revenue', by='place',
        ...     how=stats.Count()
        ... )
        >>> agg = compose.TransformerUnion(mean, count)

        We can now update each transformer and obtain their output with a single function call:

        >>> from pprint import pprint
        >>> for x in X:
        ...     agg = agg.fit_one(x)
        ...     pprint(agg.transform_one(x))
        {'revenue_count_by_place': 1, 'revenue_mean_by_place': 42.0}
        {'revenue_count_by_place': 1, 'revenue_mean_by_place': 16.0}
        {'revenue_count_by_place': 2, 'revenue_mean_by_place': 20.0}
        {'revenue_count_by_place': 2, 'revenue_mean_by_place': 50.0}
        {'revenue_count_by_place': 3, 'revenue_mean_by_place': 20.0}
        {'revenue_count_by_place': 3, 'revenue_mean_by_place': 50.0}

        Note that you can use the `+` operator as a shorthand notation:

        agg = mean + count

        This allows you to build complex pipelines in a very terse manner. For instance, we can
        create a pipeline that scales each feature and fits a logistic regression as so:

        >>> from creme import linear_model as lm
        >>> from creme import preprocessing as pp

        >>> model = (
        ...     (mean + count) |
        ...     pp.StandardScaler() |
        ...     lm.LogisticRegression()
        ... )

        Whice is equivalent to the following code:

        >>> model = compose.Pipeline(
        ...     compose.TransformerUnion(mean, count),
        ...     pp.StandardScaler(),
        ...     lm.LogisticRegression()
        ... )

        Note that you access any part of a `TransformerUnion` by name:

        >>> model['TransformerUnion']['Agg']
        Agg (
          on="revenue"
          by=['place']
          how=Mean ()
        )

        >>> model['TransformerUnion']['Agg1']
        Agg (
          on="revenue"
          by=['place']
          how=Count ()
        )

        You can also manually provide a name for each step:

        >>> agg = compose.TransformerUnion(
        ...     ('Mean revenue by place', mean),
        ...     ('# by place', count)
        ... )

    """

    def __init__(self, *transformers):
        self.transformers = {}
        for transformer in transformers:
            self += transformer

    def __getitem__(self, key):
        """Just for convenience."""
        return self.transformers[key]

    def __len__(self):
        """Just for convenience."""
        return len(self.transformers)

    def __str__(self):
        return ' + '.join(map(str, self.transformers.values()))

    def __repr__(self):
        return (
            'TransformerUnion (\n\t' +
            '\t'.join(',\n'.join(map(repr, self.transformers.values())).splitlines(True)) +
            '\n)'
        ).expandtabs(2)

    def _get_params(self):
        return dict(self.transformers.items())

    def _set_params(self, new_params=None):
        if new_params is None:
            new_params = {}
        return TransformerUnion(*[
            (name, new_params[name])
            if isinstance(new_params.get(name), base.Estimator) else
            (name, step._set_params(new_params.get(name, {})))
            for name, step in self.transformers.items()
        ])

    @property
    def _is_supervised(self):
        return any(t._is_supervised for t in self.transformers.values())

    def _add_step(self, transformer):
        """Adds a transformer while taking care of the input type."""

        name = None
        if isinstance(transformer, tuple):
            name, transformer = transformer

        # If the step is a function then wrap it in a FuncTransformer
        if isinstance(transformer, (types.FunctionType, types.LambdaType)):
            transformer = func.FuncTransformer(transformer)

        def infer_name(transformer):
            if isinstance(transformer, func.FuncTransformer):
                return infer_name(transformer.func)
            elif isinstance(transformer, (types.FunctionType, types.LambdaType)):
                return transformer.__name__
            elif hasattr(transformer, '__class__'):
                return transformer.__class__.__name__
            return str(transformer)

        # Infer a name if none is given
        if name is None:
            name = infer_name(transformer)

        if name in self.transformers:
            counter = 1
            while f'{name}{counter}' in self.transformers:
                counter += 1
            name = f'{name}{counter}'

        # Store the transformer
        self.transformers[name] = transformer

        return self

    def __add__(self, other):
        return self._add_step(other)

    def fit_one(self, x, y=None):
        """Update each transformer.

        Parameters:
            x: Features.
            y: An optional target, this is expected to be provided if at least one of the
                transformers is supervised (i.e. it inherits from `base.SupervisedTransformer`).

        """
        for t in self.transformers.values():
            if isinstance(t, base.SupervisedTransformer):
                t.fit_one(x, y)
            else:
                t.fit_one(x)
        return self

    def transform_one(self, x):
        """Passes the data through each transformer and packs the results together."""
        return dict(collections.ChainMap(*(
            t.transform_one(x)
            for t in self.transformers.values()
        )))



import abc

from creme import base


class Transformer(base.Estimator):
    """A transformer."""

    @property
    def _is_supervised(self):
        return False

    def fit_one(self, x: dict) -> 'Transformer':
        """Update with a set of features `x`.

        A lot of transformers don't actually have to do anything during the `fit_one` step
        because they are stateless. For this reason the default behavior of this function is to do
        nothing. Transformers that however do something during the `fit_one` can override this
        method.

        Parameters:
            x: A dictionary of features.

        Returns:
            self

        """
        return self

    @abc.abstractmethod
    def transform_one(self, x: dict) -> dict:
        """Transform a set of features `x`.

        Parameters:
            x: A dictionary of features.

        Returns:
            dict

        """

    def __add__(self, other):
        """Fuses with another Transformer into a TransformerUnion."""
        from .. import compose
        if isinstance(other, compose.TransformerUnion):
            return other.__add__(self)
        return compose.TransformerUnion(self, other)

    def __radd__(self, other):
        """Fuses with another Transformer into a TransformerUnion."""
        from .. import compose
        if isinstance(other, compose.TransformerUnion):
            return other.__add__(self)
        return compose.TransformerUnion(other, self)

    def __mul__(self, feature):
        """Creates a Grouper."""
        from .. import compose
        return compose.Grouper(transformer=self, by=feature)

    def __rmul__(self, feature):
        """Creates a Grouper."""
        from .. import compose
        return compose.Grouper(transformer=self, by=feature)


class SupervisedTransformer(Transformer):

    @property
    def _is_supervised(self):
        return True

    def fit_one(self, x: dict, y: base.typing.Target) -> 'SupervisedTransformer':
        """Update with a set of features `x` and a target `y`.

        Parameters:
            x: A dictionary of features.

        Returns:
            self

        """
        return self



from . import base
from . import cov
from . import var


class PearsonCorr(base.Bivariate):
    """Online Pearson correlation.

    Parameters:
        ddof: Delta Degrees of Freedom.

    Attributes:
        var_x (stats.Var): Running variance of `x`.
        var_y (stats.Var): Running variance of `y`.
        cov_xy (stats.Cov): Running covariance of `x` and `y`.

    Example:

        >>> from creme import stats

        >>> x = [0, 0, 0, 1, 1, 1, 1]
        >>> y = [0, 1, 2, 3, 4, 5, 6]

        >>> pearson = stats.PearsonCorr()

        >>> for xi, yi in zip(x, y):
        ...     print(pearson.update(xi, yi).get())
        0
        0
        0
        0.774596
        0.866025
        0.878310
        0.866025

    """

    def __init__(self, ddof=1):
        self.var_x = var.Var(ddof=ddof)
        self.var_y = var.Var(ddof=ddof)
        self.cov_xy = cov.Cov(ddof=ddof)

    @property
    def ddof(self):
        return self.cov_xy.ddof

    def update(self, x, y):
        self.var_x.update(x)
        self.var_y.update(y)
        self.cov_xy.update(x, y)
        return self

    def get(self):
        var_x = self.var_x.get()
        var_y = self.var_y.get()
        if var_x and var_y:
            return self.cov_xy.get() / (var_x * var_y) ** .5
        return 0


class RollingPearsonCorr(base.Bivariate):
    """Rolling Pearson correlation.

    Parameters:
        window_size: Amount of samples over which to compute the correlation.
        ddof: Delta Degrees of Freedom.

    Attributes:
        var_x (stats.Var): Running variance of `x`.
        var_y (stats.Var): Running variance of `y`.
        cov_xy (stats.Cov): Running covariance of `x` and `y`.

    Example:

        >>> from creme import stats

        >>> x = [0, 0, 0, 1, 1, 1, 1]
        >>> y = [0, 1, 2, 3, 4, 5, 6]

        >>> pearson = stats.RollingPearsonCorr(window_size=4)

        >>> for xi, yi in zip(x, y):
        ...     print(pearson.update(xi, yi).get())
        0
        0
        0
        0.7745966692414834
        0.894427190999916
        0.7745966692414834
        0

    """

    def __init__(self, window_size, ddof=1):
        self.var_x = var.RollingVar(window_size=window_size, ddof=ddof)
        self.var_y = var.RollingVar(window_size=window_size, ddof=ddof)
        self.cov_xy = cov.RollingCov(window_size=window_size, ddof=ddof)

    @property
    def window_size(self):
        return self.cov_xy.window_size

    def update(self, x, y):
        self.var_x.update(x)
        self.var_y.update(y)
        self.cov_xy.update(x, y)
        return self

    def get(self):
        var_x = self.var_x.get()
        var_y = self.var_y.get()
        if var_x and var_y:
            return self.cov_xy.get() / (var_x * var_y) ** 0.5
        return 0



import math

from creme import proba
from creme.proba.base import ContinuousDistribution

from .. import base


class Branch(base.Branch):

    def update(self, x, y):
        if self.split(x):
            self.left = self.left.update(x, y)
            return self
        self.right = self.right.update(x, y)
        return self


class Leaf(base.Leaf):

    def __init__(self, depth, tree, target_dist):
        self.depth = depth
        self.tree = tree
        self.target_dist = target_dist
        self.n_samples = 0
        self.split_enums = {}

    @property
    def n_classes(self):
        """The number of observed classes."""
        if isinstance(self.target_dist, ContinuousDistribution):
            raise AttributeError('The target is continuous, hence there are not classes')
        return len(self.target_dist)

    @property
    def is_pure(self):
        try:
            return self.n_classes < 2
        except AttributeError:
            return self.n_samples > 1

    @property
    def hoeffding_bound(self):
        """Returns the current Hoeffding bound."""
        R = math.log(self.n_classes)
        n = self.n_samples
        δ = self.tree.confidence
        return math.sqrt(R ** 2 * math.log(1 / δ) / (2 * n))

    def update(self, x, y):

        # Update the class counts
        self.target_dist.update(y)
        self.n_samples += 1

        # Update the sufficient statistics of each feature's split searcher
        for i, xi in x.items():
            try:
                ss = self.split_enums[i]
            except KeyError:
                ss = self.split_enums[i] = self.tree._get_split_enum(value=xi)
            ss.update(x=xi, y=y)

        # Check if splitting is authorized or not
        if (
            self.depth >= self.tree.max_depth or
            self.is_pure or
            self.n_samples % self.tree.patience != 0
        ):
            return self

        # Search for the best split given the current information
        top_2_diff, split = self.find_best_split()
        if not split:
            return self

        # Calculate the Hoeffding bound
        ε = self.hoeffding_bound

        if top_2_diff > ε or ε < self.tree.tie_threshold:
            return Branch(
                split=split,
                left=Leaf(depth=self.depth + 1, tree=self.tree, target_dist=self.tree._make_leaf_dist()),
                right=Leaf(depth=self.depth + 1, tree=self.tree, target_dist=self.tree._make_leaf_dist()),
                tree=self.tree,
                target_dist=self.target_dist,
                n_samples=self.n_samples
            )
        return self

    def find_best_split(self):
        """Returns the best potential split."""

        current_impurity = self.tree.criterion_func(dist=self.target_dist)
        best_gain = -math.inf
        second_best_gain = -math.inf
        best_split = None

        # For each feature
        for feature_name, split_enum in self.split_enums.items():

            # For each candidate split
            for how, at, l_dist, r_dist in split_enum.enumerate_splits(target_dist=self.target_dist):

                # Ignore the split if it results in a new leaf with not enough samples
                if (
                    l_dist.n_samples < self.tree.min_child_samples or
                    r_dist.n_samples < self.tree.min_child_samples
                ):
                    continue

                # Compute the decrease in impurity brought by the split
                left_impurity = self.tree.criterion_func(dist=l_dist)
                right_impurity = self.tree.criterion_func(dist=r_dist)
                impurity = l_dist.n_samples * left_impurity + r_dist.n_samples * right_impurity
                impurity /= l_dist.n_samples + r_dist.n_samples
                gain = current_impurity - impurity

                # Ignore the split if the gain in impurity is too low
                if gain < self.tree.min_split_gain:
                    continue

                # Check if the gain brought by the candidate split is better than the current best
                if gain > best_gain:
                    best_gain, second_best_gain = gain, best_gain
                    best_split = base.Split(on=feature_name, how=how, at=at)
                elif gain > second_best_gain:
                    second_best_gain = gain

        return best_gain - second_best_gain, best_split



import collections
import math

from creme import base
from creme import utils


__all__ = ['ALMAClassifier']


class ALMAClassifier(base.BinaryClassifier):
    """Approximate Large Margin Algorithm (ALMA).

    Parameters:
        p
        alpha
        B
        C

    Attributes:
        w (collections.defaultdict): The current weights.
        k (int): The number of instances seen during training.

    Example:

        >>> from creme import datasets
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import preprocessing

        >>> X_y = datasets.Phishing()

        >>> model = (
        ...     preprocessing.StandardScaler() |
        ...     linear_model.ALMAClassifier()
        ... )

        >>> metric = metrics.Accuracy()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        Accuracy: 82.64%

    References:
        1. [Gentile, Claudio. "A new approximate maximal margin classification algorithm." Journal of Machine Learning Research 2.Dec (2001): 213-242](http://www.jmlr.org/papers/volume2/gentile01a/gentile01a.pdf)

    """

    def __init__(self, p=2, alpha=.9, B=1 / .9, C=2 ** .5):
        self.p = p
        self.alpha = alpha
        self.B = B
        self.C = C
        self.w = collections.defaultdict(float)
        self.k = 1

    def _raw_dot(self, x):
        return utils.math.dot(x, self.w)

    def predict_proba_one(self, x):
        yp = utils.math.sigmoid(self._raw_dot(x))
        return {False: 1 - yp, True: yp}

    def fit_one(self, x, y):

        # Convert 0 to -1
        y = int(y or -1)

        gamma = self.B * math.sqrt(self.p - 1) / math.sqrt(self.k)

        if y * self._raw_dot(x) < (1 - self.alpha) * gamma:

            eta = self.C / (math.sqrt(self.p - 1) * math.sqrt(self.k))

            for i, xi in x.items():
                self.w[i] += eta * y * xi

            norm = utils.math.norm(self.w, order=self.p)

            for i in x:
                self.w[i] /= max(1, norm)

            self.k += 1

        return self



import collections

from . import base


__all__ = ['AdaGrad']


class AdaGrad(base.Optimizer):
    """AdaGrad optimizer.

    Parameters:
        lr
        eps

    Attributes:
        g2 (collections.defaultdict)

    Example:

        >>> from creme import datasets
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import optim
        >>> from creme import preprocessing

        >>> X_y = datasets.Phishing()
        >>> optimizer = optim.AdaGrad()
        >>> model = (
        ...     preprocessing.StandardScaler() |
        ...     linear_model.LogisticRegression(optimizer)
        ... )
        >>> metric = metrics.F1()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        F1: 0.880143

    References:
        1. [Duchi, J., Hazan, E. and Singer, Y., 2011. Adaptive subgradient methods for online learning and stochastic optimization. Journal of machine learning research, 12(Jul), pp.2121-2159.](http://www.jmlr.org/papers/volume12/duchi11a/duchi11a.pdf)

    """

    def __init__(self, lr=.1, eps=1e-8):
        super().__init__(lr)
        self.eps = eps
        self.g2 = collections.defaultdict(float)

    def _update_after_pred(self, w, g):

        for i, gi in g.items():
            self.g2[i] += gi ** 2
            w[i] -= self.learning_rate / (self.g2[i] + self.eps) ** 0.5 * gi

        return w



"""Helper functions for making things readable by humans."""
import inspect
import math
import types
import typing


__all__ = ['humanize_bytes', 'print_table']


def print_table(headers: typing.List[str], columns: typing.List[typing.List[str]],
                order: typing.List[int] = None):
    """Pretty-prints a table.

    Parameters:
        headers: The column names.
        columns: The column values.
        order: Order in which to print the column the values. Defaults to the order in which the
            values are given.

    """

    # Check inputs
    if len(headers) != len(columns):
        raise ValueError('there must be as many headers as columns')

    if len(set(map(len, columns))) > 1:
        raise ValueError('all the columns must be of the same length')

    # Determine the width of each column based on the maximum length of it's elements
    col_widths = [
        max(*map(len, col), len(header))
        for header, col in zip(headers, columns)
    ]

    # Make a template to print out rows one by one
    row_format = ' '.join(['{:' + str(width + 2) + 's}' for width in col_widths])

    # Determine the order in which to print the column values
    if order is None:
        order = range(len(columns[0]))

    # Build the table
    table = (
        row_format.format(*headers) + '\n' +
        '\n'.join((
            row_format.format(*[
                col[i].rjust(width)
                for col, width in zip(columns, col_widths)
            ])
            for i in order
        ))
    )

    return table


def humanize_bytes(n_bytes):
    """Returns a human-friendly byte size."""
    suffixes = ['B', 'KB', 'MB', 'GB', 'TB', 'PB']
    human = n_bytes
    rank = 0
    if n_bytes != 0:
        rank = int((math.log10(n_bytes)) / 3)
        rank = min(rank, len(suffixes) - 1)
        human = n_bytes / (1024.0 ** rank)
    f = ('%.2f' % human).rstrip('0').rstrip('.')
    return '%s %s' % (f, suffixes[rank])



import glob
import os

from . import base


class MaliciousURL(base.RemoteDataset):
    """Malicious URLs dataset.

    This dataset contains features about URLs that are classified as malicious or not.

    References:
        1. [Detecting Malicious URLs](http://www.sysnet.ucsd.edu/projects/url/)
        2. [Identifying Suspicious URLs: An Application of Large-Scale Online Learning](http://cseweb.ucsd.edu/~jtma/papers/url-icml2009.pdf)

    """

    def __init__(self):
        super().__init__(
            n_samples=2_396_130,
            n_features=3_231_961,
            task=base.BINARY_CLF,
            url='http://www.sysnet.ucsd.edu/projects/url/url_svmlight.tar.gz',
            filename='url_svmlight',
            size=2210273352,
            sparse=True
        )

    def _iter(self):

        files = list(self.path.glob('Day*.svm'))
        files.sort(key=lambda x: int(os.path.basename(x).split('.')[0][3:]))

        def parse_libsvm_feature(f):
            k, v = f.split(':')
            return int(k), float(v)

        # There are 150 files with each one corresponding to a day
        for file in files:
            with open(file) as f:
                for line in f:
                    # Each file has the libsvm format
                    elements = line.rstrip().split(' ')
                    y = elements.pop(0) == '+1'
                    x = dict(parse_libsvm_feature(f) for f in elements)
                    yield x, y



import abc

from creme import base


class Wrapper(base.Estimator):
    """A wrapper model."""

    @abc.abstractproperty
    def _wrapped_model(self):
        """Provides access to the wrapped model."""

    @property
    def _labelloc(self):
        """Indicates location of the wrapper name when drawing pipelines."""
        return 't'  # for top

    def __str__(self):
        return f'{type(self).__name__}({self._wrapped_model})'



"""Decision trees."""
from .decision.forest import RandomForestClassifier
from .decision.tree import DecisionTreeClassifier


__all__ = [
    'DecisionTreeClassifier',
    'RandomForestClassifier'
]

__pdoc__ = {'decision': False}



import numpy as np

from . import window


class SDFT(window.Window):
    """Sliding Discrete Fourier Transform (SDFT).

    Initially, the coefficients are all equal to 0, up until enough values have been seen. A call
    to `numpy.fft.fft` is triggered once ``window_size`` values have been seen. Subsequent values
    will update the coefficients online. This is much faster than recomputing an FFT from scratch
    for every new value.

    Parameters:
        window_size (int): The size of the window.

    Attributes:
        window (window.Window): The window of values.

    Example:

        ::

            >>> from creme import utils

            >>> X = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

            >>> window_size = 5
            >>> sdft = utils.SDFT(window_size)

            >>> for i, x in enumerate(X):
            ...     sdft = sdft.update(x)
            ...
            ...     if i + 1 >= window_size:
            ...         assert np.allclose(sdft, np.fft.fft(X[i+1 - window_size:i+1]))

    References:
        1. `Jacobsen, E. and Lyons, R., 2003. The sliding DFT. IEEE Signal Processing Magazine, 20(2), pp.74-80. <https://www.comm.utoronto.ca/~dimitris/ece431/slidingdft.pdf>`_
        2. `Understanding and Implementing the Sliding DFT <https://www.dsprelated.com/showarticle/776.php>`_

    """

    def __init__(self, window_size):
        super().__init__(size=window_size)
        self.window = window.Window(size=window_size)

    def update(self, x):

        # Simply append the new value if the window isn't full yet
        if len(self.window) < self.window.size - 1:
            self.window.append(x)

        # Compute an initial FFT the first time the window is full
        elif len(self.window) == self.window.size - 1:
            self.window.append(x)
            self.extend(np.fft.fft(self.window))

        # Update the coefficients for subsequent values
        else:
            diff = x - self.window[0]
            for i in range(self.size):
                self[i] = (self[i] + diff) * np.exp(2j * np.pi * i / self.size)
            self.window.append(x)

        return self



from collections import defaultdict
import functools

import numpy as np
from scipy import special
from scipy import ndimage

from .. import base
from ..feature_extraction import vectorize


__all__ = ['LDA']


class LDA(base.Transformer, vectorize.VectorizerMixin):
    """Online Latent Dirichlet Allocation with Infinite Vocabulary.

    Latent Dirichlet allocation (LDA) is a probabilistic approach for exploring topics in document
    collections. The key advantage of this variant is that it assumes an infinite vocabulary,
    meaning that the set of tokens does not have to known in advance, as opposed to the
    [implementation from sklearn](https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.LatentDirichletAllocation.html)
    The results produced by this implementation are identical to those from [the original
    implementation](https://github.com/kzhai/PyInfVoc) proposed by the method's authors.

    Parameters:
        n_components: Number of topics of the latent Drichlet allocation.
        number_of_documents: Estimated number of documents.
        on: The name of the feature that contains the text to vectorize. If `sNone`s, then
            the input is treated as a document instead of a set of features.
        strip_accents: Whether or not to strip accent characters.
        lowercase: Whether or not to convert all characters to lowercase.
        preprocessor: Override the preprocessing step while preserving the tokenizing
            and n-grams generation steps.
        tokenizer: The function used to convert preprocessed text into a `dict` of
            tokens. A default one is used if it is not provided by the user.
        ngram_range: The lower and upper boundary of the range n-grams to be
            extracted. All values of n such that `smin_n <= n <= max_n`s will be used. For example
            an `sngram_range`s of `s(1, 1)`s means only unigrams, `s(1, 2)`s means unigrams and
            bigrams, and `s(2, 2)`s means only bigrams. Only works if `stokenizer`s is not set to
            `sFalse`s.
        alpha_theta: Hyper-parameter of the Dirichlet distribution of topics.
        alpha_beta: Hyper-parameter of the Dirichlet process of distribution over words.
        tau: Learning inertia to prevent premature convergence.
        kappa: The learning rate kappa controls how quickly new parameters estimates
            replace the old ones. kappa ∈ (0.5, 1] is required for convergence.
        vocab_prune_interval: Interval at which to refresh the words topics distribution.
        number_of_samples: Number of iteration to computes documents topics distribution.
        burn_in_sweeps: Number of iteration necessaries while analyzing a document
            before updating document topics distribution.
        maximum_size_vocabulary: Maximum size of the stored vocabulary.

    Attributes:
        counter (int): The current number of observed documents.
        truncation_size_prime (int): Number of distincts words stored in the vocabulary. Updated
            before processing a document.
        truncation_size (int) : Number of distincts words stored in the vocabulary. Updated after
            processing a document.
        word_to_index (dict): Words as keys and indexes as values.
        index_to_word (dict): Indexes as keys and words as values.
        nu_1 (dict): Weights of the words. Component of the variational inference.
        nu_2 (dict): Weights of the words. Component of the variational inference.

    Example:

        >>> from creme import decomposition
        >>> import numpy as np

        >>> np.random.seed(42)

        >>> X = [
        ...    'weather cold',
        ...    'weather hot dry',
        ...    'weather cold rainy',
        ...    'weather hot',
        ...    'weather cold humid',
        ... ]

        >>> online_lda = decomposition.LDA(n_components=2, number_of_documents=60)
        >>> for x in X:
        ...     print(online_lda.fit_transform_one(x))
        {0: 0.5, 1: 2.5}
        {0: 3.5, 1: 0.5}
        {0: 0.5, 1: 3.5}
        {0: 1.5, 1: 1.5}
        {0: 2.5, 1: 1.5}

    References:
        1. [Zhai, K. and Boyd-Graber, J., 2013, February. Online latent Dirichlet allocation with infinite vocabulary. In International Conference on Machine Learning (pp. 561-569).](http://proceedings.mlr.press/v28/zhai13.pdf)
        2. [PyInfVoc on GitHub](https://github.com/kzhai/PyInfVoc)

    """

    def __init__(self, n_components=10, number_of_documents=1e6, on=None, strip_accents=True,
                 lowercase=True, preprocessor=None, tokenizer=None, ngram_range=(1, 1),
                 alpha_theta=0.5, alpha_beta=100., tau=64., kappa=0.75, vocab_prune_interval=10,
                 number_of_samples=10, ranking_smooth_factor=1e-12, burn_in_sweeps=5,
                 maximum_size_vocabulary=4000):

        # Initialize the VectorizerMixin part
        super().__init__(
            on=on,
            strip_accents=strip_accents,
            lowercase=lowercase,
            preprocessor=preprocessor,
            tokenizer=tokenizer,
            ngram_range=ngram_range
        )

        self.n_components = n_components
        self.number_of_documents = number_of_documents
        self.alpha_theta = alpha_theta
        self.alpha_beta = alpha_beta
        self.tau = tau
        self.kappa = kappa
        self.vocab_prune_interval = vocab_prune_interval
        self.number_of_samples = number_of_samples
        self.ranking_smooth_factor = ranking_smooth_factor
        self.burn_in_sweeps = burn_in_sweeps
        self.maximum_size_vocabulary = maximum_size_vocabulary

        self.counter = 0
        self.truncation_size_prime = 1
        self.truncation_size = 1

        self.word_to_index = {}
        self.index_to_word = {}

        self.nu_1 = defaultdict(functools.partial(np.ones, 1))
        self.nu_2 = defaultdict(functools.partial(np.array, [self.alpha_beta]))

        for topic in range(self.n_components):
            self.nu_1[topic] = np.ones(1)
            self.nu_2[topic] = np.array([self.alpha_beta])

    def fit_transform_one(self, x):
        """Equivalent to `slda.fit_one(x).transform_one(x)`s, but faster.

        Parameters:
            x (`dict` or `str`): Input features containing a text document, or text document.

        Returns
            dict: Components of the input document.

        """

        # Updates number of documents:
        self.counter += 1

        # Extracts words of the document as a list of words:
        word_list = self.process_text(x)

        # Update words indexes:
        self._update_indexes(word_list=word_list)

        # Replace the words by their index:
        words_indexes_list = [self.word_to_index[word] for word in word_list]

        # Sample empirical topic assignment:
        statistics, batch_document_topic_distribution = self._compute_statistics_components(
            words_indexes_list
        )

        # Online variational inference
        self._update_weights(statistics=statistics)

        if self.counter % self.vocab_prune_interval == 0:
            self._prune_vocabulary()

        return dict(enumerate(batch_document_topic_distribution))

    def fit_one(self, x):
        """Updates the LDA.

        Tokenizes the document, then updates the word indexes, and then updates the online
        variational inference.

        Args:
            x (`dict` or `str`): Input features containing a text document, or text document.

        Returns
            dict: Components of the input document.

        """
        # Updates number of documents:
        self.counter += 1

        # Extracts words of the document as a list of words:
        word_list = self.process_text(x)

        # Update words indexes:
        self._update_indexes(word_list=word_list)

        # Replace the words by their index:
        words_indexes_list = [self.word_to_index[word] for word in word_list]

        # Sample empirical topic assignment:
        statistics, _ = self._compute_statistics_components(words_indexes_list)

        # Oline variational inference
        self._update_weights(statistics=statistics)

        if self.counter % self.vocab_prune_interval == 0:
            self._prune_vocabulary()

        return self

    def transform_one(self, x):
        """Returns document topics.

        Tokenizes the document, then updates the word indexes, and then assigns topics to the
        document.

        Parameters:
            x (`dict` or `str`): Input features containing a text document, or text document.

        Returns
            dict: Components of the input document.

        """
        # Extracts words of the document as a list of words:
        word_list = self.process_text(x)

        # Update words indexes:
        self._update_indexes(word_list=word_list)

        # Replace the words by their index:
        words_indexes_list = [self.word_to_index[word] for word in word_list]

        # Sample empirical topic assignment:
        _, components = self._compute_statistics_components(words_indexes_list)

        return dict(enumerate(components))

    def _update_indexes(self, word_list):
        """
        Adds the words of the document to the index if they are not part of the current vocabulary.
        Updates of the number of distinct words seen.

        Args:
            word_list (list): Content of the document as a list of words.

        Returns:
            None
        """
        for word in word_list:
            if word not in self.word_to_index:
                new_index = len(self.word_to_index) + 1
                self.word_to_index[word] = new_index
                self.index_to_word[new_index] = word
                self.truncation_size_prime += 1

    @classmethod
    def _compute_weights(cls, n_components, nu_1, nu_2):
        """
        Calculates the vocabulary weighting according to the word distribution present in the
        vocabulary. The Psi function is the logarithmic derivative of the gamma function.

        Args:
            n_components (int): Number of topics.
            nu_1 (dict): Weights of the words of the vocabulary.
            nu_2 (dict): Weights of the words of the vocabulary.

        Returns:
            Tuple[dict, dict]: Weights of the words of the current vocabulary.
        """
        exp_weights = {}
        exp_oov_weights = {}

        for topic in range(n_components):

            psi_nu_1 = special.psi(nu_1[topic])
            psi_nu_2 = special.psi(nu_2[topic])

            psi_nu_1_nu_2 = special.psi(nu_1[topic] + nu_2[topic])

            psi_nu_1_nu_2_minus_psi_nu_2 = np.cumsum([psi_nu_2 - psi_nu_1_nu_2], axis=1)

            exp_oov_weights[topic] = np.exp(psi_nu_1_nu_2_minus_psi_nu_2[0][-1])

            psi_nu_1_nu_2_minus_psi_nu_2 = ndimage.interpolation.shift(
                input=psi_nu_1_nu_2_minus_psi_nu_2[0],
                shift=1,
                cval=0
            )

            exp_weights[topic] = np.exp(psi_nu_1 - psi_nu_1_nu_2 + psi_nu_1_nu_2_minus_psi_nu_2)

        return exp_weights, exp_oov_weights

    def _update_weights(self, statistics):
        """
        Learns documents and word representations. Calculates the variational approximation.

        Args:
            statistics (defaultdict): Weights associated to the words.

        Returns:
            None
        """
        reverse_cumulated_phi = {}

        for k in range(self.n_components):

            reverse_cumulated_phi[k] = ndimage.interpolation.shift(
                input=statistics[k],
                shift=-1,
                cval=0
            )

            reverse_cumulated_phi[k] = np.flip(reverse_cumulated_phi[k])
            reverse_cumulated_phi[k] = np.cumsum(reverse_cumulated_phi[k])
            reverse_cumulated_phi[k] = np.flip(reverse_cumulated_phi[k])

        # Epsilon will be between 0 and 1.
        # Epsilon value says how much to weight the information we got from this document.
        self.epsilon = (self.tau + self.counter) ** -self.kappa

        for k in range(self.n_components):

            if self.truncation_size < self.truncation_size_prime:

                difference_truncation = self.truncation_size_prime - self.truncation_size

                self.nu_1[k] = np.append(self.nu_1[k], np.ones(difference_truncation))
                self.nu_2[k] = np.append(self.nu_2[k], np.ones(difference_truncation))

            # Variational Approximation
            self.nu_1[k] += (self.epsilon * (self.number_of_documents *
                                             np.array(statistics[k]) + 1 - self.nu_1[k]))

            self.nu_2[k] += (self.epsilon * (self.alpha_beta + self.number_of_documents *
                                             np.array(reverse_cumulated_phi[k]) - self.nu_2[k]))

        self.truncation_size = self.truncation_size_prime

    def _compute_statistics_components(self, words_indexes_list):
        """
        Extract latent variables from the document and words.

        Args:
            words_indexes_list (list): Ids of the words of the input document.

        Returns:
            Tuple[dict, dict]: Computed statistics over the words. Document reprensetation across
            topics.
        """
        statistics = defaultdict(lambda: np.zeros(self.truncation_size_prime))

        exp_weights, exp_oov_weights = self._compute_weights(
            n_components=self.n_components,
            nu_1=self.nu_1,
            nu_2=self.nu_2,
        )

        size_vocab = len(words_indexes_list)

        phi = np.random.random((self.n_components, size_vocab))

        phi = phi / np.sum(phi, axis=0)

        phi_sum = np.sum(phi, axis=1)

        for sample_index in range(self.number_of_samples):

            for word_index in range(size_vocab):

                phi_sum -= phi[:, word_index]
                phi_sum = phi_sum.clip(min=0)
                temp_phi = phi_sum + self.alpha_theta

                for k in range(self.n_components):

                    if words_indexes_list[word_index] >= self.truncation_size:

                        temp_phi[k] *= exp_oov_weights[k]
                    else:
                        temp_phi[k] *= exp_weights[k][words_indexes_list[word_index]]

                # Normalize document topic distribution before applying multinomial distribution:
                temp_phi /= temp_phi.sum()

                # Sample a topic based a given probability distribution:
                temp_phi = np.random.multinomial(1, temp_phi)

                phi[:, word_index] = temp_phi

                phi_sum += temp_phi

                if sample_index >= self.burn_in_sweeps:

                    for k in range(self.n_components):

                        index = words_indexes_list[word_index]

                        statistics[k][index] += temp_phi[k]

        document_topic_distribution = self.alpha_theta + phi_sum

        for k in range(self.n_components):

            statistics[k] /= (self.number_of_samples - self.burn_in_sweeps)

        return statistics, document_topic_distribution

    def _prune_vocabulary(self):
        """
        Reduces the size of the index exceeds the maximum size.

        Args:
            None

        Returns:
            None
        """
        if self.nu_1[0].shape[0] > self.maximum_size_vocabulary:

            for topic in range(self.n_components):
                # Updates words latent variables
                self.nu_1[topic] = self.nu_1[topic][:self.maximum_size_vocabulary]
                self.nu_2[topic] = self.nu_2[topic][:self.maximum_size_vocabulary]

            new_word_to_index = {}
            new_index_to_word = {}

            for index in range(1, self.maximum_size_vocabulary):
                # Updates words indexes
                word = self.index_to_word[index]
                new_word_to_index[word] = index
                new_index_to_word[index] = word

            self.word_to_index = new_word_to_index
            self.index_to_word = new_index_to_word

            self.truncation_size = self.nu_1[0].shape[0]
            self.truncation_size_prime = self.truncation_size



from . import base


__all__ = ['RegressionMultiOutput']


class RegressionMultiOutput(base.MultiOutputRegressionMetric, base.WrapperMetric):
    """Wrapper for multi-output regression.

    This wraps a regression metric to make it compatible with multi-output regression tasks. The
    value of each output will be fed sequentially to the `get` method of the provided metric.

    Parameters:
        metric: The regression metric to evaluate with each output.

    """

    def __init__(self, metric: 'base.RegressionMetric'):
        self._metric = metric

    @property
    def metric(self):
        return self._metric

    def update(self, y_true, y_pred, sample_weight=1.):
        for i in y_true:
            self.metric.update(y_true[i], y_pred[i], sample_weight)
        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        for i in y_true:
            self.metric.revert(y_true[i], y_pred[i], sample_weight)
        return self

    def get(self):
        return self.metric.get()



import collections
import typing

from . import base


__all__ = ['Multinomial']


class Multinomial(base.DiscreteDistribution):
    """Multinomial distribution for categorical data.

    Parameters:
        events: An optional list of events that already occurred.

    Example:

        >>> from creme import proba

        >>> p = proba.Multinomial(['green'] * 3)
        >>> p = p.update('red')

        >>> p.pmf('red')
        0.25

        >>> p.update('red').update('red').pmf('green')
        0.5

    """

    def __init__(self, events: typing.Union[dict, list] = None):
        self.counts: typing.Counter[typing.Any] = collections.Counter(events)
        self._n = sum(self.counts.values())

    @property
    def n_samples(self):
        return self._n

    def __iter__(self):
        return iter(self.counts)

    def __len__(self):
        return len(self.counts)

    @property
    def mode(self):
        return self.counts.most_common(1)[0][0]

    def update(self, x):
        self.counts.update([x])
        self._n += 1
        return self

    def pmf(self, x):
        try:
            return self.counts[x] / self._n
        except ZeroDivisionError:
            return 0.

    def __str__(self):
        return '\n'.join(f'P({c}) = {self.pmf(c):.3f}' for c in self.counts.most_common())



from creme import stream

from . import base


class SMTP(base.RemoteDataset):
    """SMTP dataset from the KDD 1999 cup.

    The goal is to predict whether or not an SMTP connection is anomalous or not. The dataset only
    contains 2,211 (0.4%) positive labels.

    References:
        1. [SMTP (KDDCUP99) dataset](http://odds.cs.stonybrook.edu/smtp-kddcup99-dataset/)

    """

    def __init__(self):
        super().__init__(
            n_samples=95_156,
            n_features=3,
            task=base.BINARY_CLF,
            url='https://maxhalford.github.io/files/datasets/smtp.zip',
            size=5484982,
            filename='smtp.csv'
        )

    def _iter(self):
        return stream.iter_csv(
            self.path,
            target='service',
            converters={
                'duration': float,
                'src_bytes': float,
                'dst_bytes': float,
                'service': int
            }
        )



"""Missing data imputation.

As a convention, missing values are indicated with an explicit `None` value. In other words, they
are present in the dictionary of features. For instance, in the following example, the feature
`shop` has a missing value:

```python
x = {'country': 'Sweden', 'shop': None}
```

"""
from .previous import PreviousImputer
from .stat import StatImputer


__all__ = [
    'PreviousImputer',
    'StatImputer'
]



import json
import os

import pytest
import sqlalchemy as sql
from sqlalchemy import orm
import zipfile

from creme import stream


@pytest.fixture
def pokedb():
    engine = sql.create_engine('sqlite://')  # in-memory

    # Load the fixtures
    here = os.path.dirname(os.path.realpath(__file__))
    with zipfile.ZipFile(os.path.join(here, 'pokedb.zip')) as z:
        pokemons = json.loads(z.read('pokemons.json'))
        types = json.loads(z.read('types.json'))
        pokemon_types = json.loads(z.read('pokemon_types.json'))

    # Define the tables
    metadata = sql.MetaData()

    t_pokemons = sql.Table('pokemons', metadata,
        sql.Column('id', sql.Integer, primary_key=True),
        sql.Column('name', sql.String),
        sql.Column('HP', sql.Integer),
        sql.Column('Attack', sql.Integer),
        sql.Column('Defense', sql.Integer),
        sql.Column('Sp. Attack', sql.Integer),
        sql.Column('Sp. Defense', sql.Integer),
        sql.Column('Speed', sql.Integer)
    )

    t_types = sql.Table('types', metadata,
        sql.Column('id', sql.Integer, primary_key=True),
        sql.Column('name', sql.String)
    )

    t_pokemon_types = sql.Table('pokemon_types', metadata,
        sql.Column('pokemon_id', sql.Integer, primary_key=True),
        sql.Column('type_id', sql.Integer, primary_key=True),
        sql.Column('no', sql.Integer, primary_key=True)
    )

    # Create the tables
    metadata.create_all(engine)

    # Insert the fixtures
    with engine.connect() as conn:
        conn.execute(t_pokemons.insert(), pokemons)
        conn.execute(t_types.insert(), types)
        conn.execute(t_pokemon_types.insert(), pokemon_types)

    return engine


def test_iter_sql(pokedb):

    with pokedb.connect() as conn:
        X_y = stream.iter_sql(query='SELECT * FROM pokemons;', conn=conn)
        x, y = next(X_y)
        assert x['name'] == 'Bulbasaur'
        assert y == None

    # This raises an exception because the resource is closed...
    with pytest.raises(sql.exc.StatementError):
        for x, y in stream.iter_sql(query='SELECT * FROM pokemons;', conn=conn):
            pass

    # ... and yet we can still stream over the results because SQLAlchemy prefetches them
    x, y = next(X_y)
    assert x['name'] == 'Ivysaur'

    # The Pokedex from generation 1 contains 151 pokemons, and we've already seen 2 of them
    assert sum(1 for _ in X_y) == 149

    # Check that the stream is depleted
    assert sum(1 for _ in X_y) == 0


def test_iter_sql_join(pokedb):

    query = '''
        SELECT
            p.name,
            t1.name AS type_1,
            t2.name AS type_2
        FROM
            pokemons p,
            pokemon_types pt1,
            pokemon_types pt2,
            types t1,
            types t2
        WHERE
            pt1.no = 1 AND
            pt1.pokemon_id = p.id AND
            pt1.type_id = t1.id AND

            pt2.no = 2 AND
            pt2.pokemon_id = p.id AND
            pt2.type_id = t2.id;
    '''

    with pokedb.connect() as conn:
        X_y = stream.iter_sql(query=query, conn=conn)
        x, y = next(X_y)
        assert x['name'] == 'Bulbasaur'
        assert x['type_1'] == 'Grass'
        assert x['type_2'] == 'Poison'



from . import base
from . import maximum
from . import minimum


class PeakToPeak(base.Univariate):
    """Running peak to peak (max - min).

    Attributes:
        max (stats.Max): The running max.
        min (stats.Min): The running min.
        p2p (float): The running peak to peak.

    Example:

        >>> from creme import stats

        >>> X = [1, -4, 3, -2, 2, 4]
        >>> ptp = stats.PeakToPeak()
        >>> for x in X:
        ...     print(ptp.update(x).get())
        0
        5
        7
        7
        7
        8

    """

    def __init__(self):
        self.max = maximum.Max()
        self.min = minimum.Min()

    @property
    def name(self):
        return 'ptp'

    def update(self, x):
        self.max.update(x)
        self.min.update(x)
        return self

    def get(self):
        return self.max.get() - self.min.get()


class RollingPeakToPeak(base.RollingUnivariate):
    """Running peak to peak (max - min) over a window.

    Parameters:
        window_size: Size of the rolling window.

    Attributes:
        max (stars.RollingMax): The running rolling max.
        min (stars.RollingMin): The running rolling min.

    Example:

        >>> from creme import stats

        >>> X = [1, -4, 3, -2, 2, 1]
        >>> ptp = stats.RollingPeakToPeak(window_size=2)
        >>> for x in X:
        ...     print(ptp.update(x).get())
        0
        5
        7
        5
        4
        1

    """

    def __init__(self, window_size: int):
        self.max = maximum.RollingMax(window_size)
        self.min = minimum.RollingMin(window_size)

    @property
    def window_size(self):
        return self.max.window_size

    def update(self, x):
        self.max.update(x)
        self.min.update(x)
        return self

    def get(self):
        return self.max.get() - self.min.get()



import math

from creme import compose
from creme import feature_extraction
from creme import naive_bayes


def test_predict_class_given_unseen_features():
    model = compose.Pipeline(
        ('tokenize', feature_extraction.BagOfWords()),
        ('nb', naive_bayes.MultinomialNB(alpha=1))
    )

    docs = [
        ('cloudy cold', 0),
        ('sunny warm', 1),
    ]

    for sentence, label in docs:
        model = model.fit_one(sentence, label)

    # Assert model parameters needed to calculate the likelihoods
    assert model['nb'].n_terms == 4
    assert model['nb'].class_totals[0] == 2
    assert model['nb'].class_totals[1] == 2

    # Given new, unseen text, predict the label
    text = 'new word'
    tokens = model['tokenize'].transform_one(text)
    cp = model['nb'].p_feature_given_class

    # P(new|0)
    #   = (N_new_0 + 1) / N_0 + N_terms)
    #   = (0 + 1) / (model['nb'].class_totals[0] + model['nb'].n_terms)
    assert cp('new', 0) == (0 + 1) / (2 + 4)

    # Since class_totals[0] == class_totals[1], and both words in text are new/unseen,
    # expect the class-conditional probabilities to be the same
    assert cp('new', 0) == cp('word', 0)
    assert cp('new', 0) == cp('new', 1)
    assert cp('new', 0) == cp('word', 1)

    jll = model['nb'].joint_log_likelihood(tokens)

    # Expect JLLs to be equal
    assert jll[0] == jll[1]

    # P(0|new word)
    #   = P(new|0) * P(word|0) * P(0)
    assert jll[0] == math.log(cp('new', 0) * cp('word', 0) * (1/2))

    # JLLs for both labels are the same, but 0 was the first label to be added to model['nb'].class_counts
    assert model.predict_one(text) == 0



import abc

from .. import base


class Forecaster(base.Estimator):

    def fit_one(self, y, x=None) -> 'Forecaster':
        """Updates the model.

        Parameters:
            y (float): In the litterature this is called the endogenous variable.
            x (dict): Optional additional features to learn from. In the litterature these are
                called the exogenous variables.

        """

    @abc.abstractmethod
    def forecast(self, horizon: int, xs=None) -> list:
        """Makes forecast at each step of the given horizon.

        Parameters:
            horizon (int): The number of steps ahead to forecast.
            xs (list): The set of . If given, then it's length should be equal to horizon.

        """



"""Evaluation metrics.

Note that the binary classification metrics expect the ground truths you provide them with to be
boolean values. In other words you need to pass a value of type `bool` to the `y_true` argument in
the `update` method of each binary metric. You will obtain incorrect results if instead you pass a
0 or 1 integer.

"""

from .accuracy import Accuracy
from .base import Metric
from .confusion import ConfusionMatrix
from .cross_entropy import CrossEntropy
from .fbeta import F1
from .fbeta import FBeta
from .fbeta import MacroF1
from .fbeta import MacroFBeta
from .fbeta import MicroF1
from .fbeta import MicroFBeta
from .fbeta import MultiFBeta
from .fbeta import WeightedF1
from .fbeta import WeightedFBeta
from .jaccard import Jaccard
from .log_loss import LogLoss
from .mae import MAE
from .mcc import MCC
from .mse import MSE
from .multioutput import RegressionMultiOutput
from .precision import MacroPrecision
from .precision import MicroPrecision
from .precision import Precision
from .precision import WeightedPrecision
from .recall import MacroRecall
from .recall import MicroRecall
from .recall import Recall
from .recall import WeightedRecall
from .report import ClassificationReport
from .rmse import RMSE
from .rmsle import RMSLE
from .roc_auc import ROCAUC
from .rolling import Rolling
from .smape import SMAPE
from .time_rolling import TimeRolling


__all__ = [
    'Accuracy',
    'ClassificationReport',
    'ConfusionMatrix',
    'CrossEntropy',
    'F1',
    'FBeta',
    'Jaccard',
    'LogLoss',
    'MAE',
    'MacroF1',
    'MacroFBeta',
    'MacroPrecision',
    'MacroRecall',
    'MCC',
    'Metric',
    'MicroF1',
    'MicroFBeta',
    'MicroPrecision',
    'MicroRecall',
    'MSE',
    'MultiFBeta',
    'Precision',
    'Recall',
    'RegressionMultiOutput',
    'RMSE',
    'RMSLE',
    'ROCAUC',
    'Rolling',
    'SMAPE',
    'TimeRolling',
    'WeightedF1',
    'WeightedFBeta',
    'WeightedPrecision',
    'WeightedRecall'
]



import importlib
import inspect
from urllib import request

import pytest

from creme import datasets
from . import base


def _iter_datasets():

    for variant in datasets.Insects.variant_sizes:
        yield datasets.Insects(variant=variant)

    for _, dataset in inspect.getmembers(importlib.import_module(f'creme.datasets'), inspect.isclass):
        if not issubclass(dataset, datasets.Insects):
            yield dataset()


@pytest.mark.parametrize(
    'dataset',
    [
        pytest.param(dataset, id=dataset.__class__.__name__)
        for dataset in _iter_datasets()
        if isinstance(dataset, base.RemoteDataset)
    ]
)
@pytest.mark.datasets
def test_remote_url(dataset):
    with request.urlopen(dataset.url) as r:
        assert r.status == 200


@pytest.mark.parametrize(
    'dataset',
    [
        pytest.param(dataset, id=dataset.__class__.__name__)
        for dataset in _iter_datasets()
        if isinstance(dataset, base.RemoteDataset)
    ]
)
@pytest.mark.datasets
def test_remote_size(dataset):
    if dataset.path.is_file():
        size = dataset.path.stat().st_size
    else:
        size = sum(f.stat().st_size for f in dataset.path.glob('**/*') if f.is_file())
    assert size == dataset.size


@pytest.mark.parametrize(
    'dataset',
    [
        pytest.param(dataset, id=dataset.__class__.__name__)
        for dataset in _iter_datasets()
        if not isinstance(dataset, base.SyntheticDataset)
    ]
)
@pytest.mark.datasets
def test_dimensions(dataset):
    n = 0
    for x, _ in dataset:
        if not dataset.sparse:
            assert len(x) == dataset.n_features
        n += 1
    assert n == dataset.n_samples



from creme import stream

from . import base


class TREC07(base.RemoteDataset):
    """TREC's 2007 Spam Track dataset.

    The data contains 75,419 chronologically ordered items, i.e. 3 months of emails delivered
    to a particular server in 2007. Spam messages represent 66.6% of the dataset.
    The goal is to predict whether an email is a spam or not.

    The available raw features are: sender, recipients, date, subject, body.

    References:
        1. [TREC 2007 Spam Track Overview](https://trec.nist.gov/pubs/trec16/papers/SPAM.OVERVIEW16.pdf)
        2. [Code ran to parse the dataset](https://gist.github.com/gbolmier/b6a942699aaaedec54041a32e4f34d40)

    """

    def __init__(self):
        super().__init__(
            n_samples=75_419,
            n_features=5,
            task=base.BINARY_CLF,
            url='https://maxhalford.github.io/files/datasets/trec07p.zip',
            size=144504829,
            filename='trec07p.csv'
        )

    def _iter(self):
        return stream.iter_csv(
            self.path,
            target='y',
            delimiter=',',
            quotechar='"',
            field_size_limit=1_000_000,
        )



"""Meta-models."""
from .pred_clipper import PredClipper
from .target_transform import BoxCoxRegressor
from .target_transform import TransformedTargetRegressor


__all__ = [
    'BoxCoxRegressor',
    'PredClipper',
    'TransformedTargetRegressor'
]



import collections
import typing

from creme import utils

from . import base


__all__ = ['Mode']


class Mode(base.Univariate):
    """Running mode.

    The mode is simply the most common value. An approximate mode can be computed by setting the
    number of first unique values to count.

    Parameters:
        k: Only the first `k` unique values will be included. If `k` equals -1, the exact mode is
            computed.

    Example:

        >>> from creme import stats

        >>> X = ['sunny', 'cloudy', 'cloudy', 'rainy', 'rainy', 'rainy']
        >>> mode = stats.Mode(k=2)
        >>> for x in X:
        ...     print(mode.update(x).get())
        sunny
        sunny
        cloudy
        cloudy
        cloudy
        cloudy

        >>> mode = stats.Mode(k=-1)
        >>> for x in X:
        ...     print(mode.update(x).get())
        sunny
        sunny
        cloudy
        cloudy
        cloudy
        rainy

    """

    def __init__(self, k=25):
        self.k = k
        self.counts = collections.defaultdict(int)

    @property
    def name(self):
        return 'mode'

    def update(self, x):
        if self.k == -1 or x in self.counts or len(self.counts) < self.k:
            self.counts[x] += 1
        return self

    def get(self):
        return max(self.counts, key=self.counts.get)


class RollingMode(base.RollingUnivariate, utils.Window):
    """Running mode over a window.

    The mode is the most common value.

    Parameters:
        window_size: Size of the rolling window.

    Attributes:
        counts (collections.defaultdict): Value counts.

    Example:

        >>> from creme import stats

        >>> X = ['sunny', 'sunny', 'sunny', 'rainy', 'rainy', 'rainy', 'rainy']
        >>> rolling_mode = stats.RollingMode(window_size=2)
        >>> for x in X:
        ...     print(rolling_mode.update(x).get())
        sunny
        sunny
        sunny
        sunny
        rainy
        rainy
        rainy

        >>> rolling_mode = stats.RollingMode(window_size=5)
        >>> for x in X:
        ...     print(rolling_mode.update(x).get())
        sunny
        sunny
        sunny
        sunny
        sunny
        rainy
        rainy

    """

    def __init__(self, window_size: int):
        super().__init__(size=window_size)
        self.counts: typing.DefaultDict[typing.Any, int] = collections.defaultdict(int)

    @property
    def window_size(self):
        return self.size

    def update(self, x):
        if len(self) >= self.size:

            # Subtract the counter of the last element
            first_in = self[0]
            self.counts[first_in] -= 1

            # No need to store the value if it's counter is 0
            if self.counts[first_in] == 0:
                self.counts.pop(first_in)

        self.counts[x] += 1
        super().append(x)
        return self

    def get(self):
        return max(self.counts, key=self.counts.get)



import collections
import functools
import operator
import random
import typing

from creme import base
from creme import preprocessing
from creme.tree.base import Leaf, Branch, Split


__all__ = ['HalfSpaceTrees']


def make_padded_tree(limits, height, padding, rng=random, **node_params):

    if height == 0:
        return Leaf(**node_params)

    # Randomly pick a feature
    # We weight each feature by the gap between each feature's limits
    on = rng.choices(
        population=list(limits.keys()),
        weights=[limits[i][1] - limits[i][0] for i in limits]
    )[0]

    # Pick a split point; use padding to avoid too narrow a split
    a = limits[on][0]
    b = limits[on][1]
    at = rng.uniform(a + padding * (b - a), b - padding * (b - a))

    # Build the left node
    tmp = limits[on]
    limits[on] = (tmp[0], at)
    left = make_padded_tree(limits=limits, height=height - 1, padding=padding, rng=rng, **node_params)
    limits[on] = tmp

    # Build the right node
    tmp = limits[on]
    limits[on] = (at, tmp[1])
    right = make_padded_tree(limits=limits, height=height - 1, padding=padding, rng=rng, **node_params)
    limits[on] = tmp

    split = Split(on=on, how=operator.lt, at=at)
    return Branch(split=split, left=left, right=right, **node_params)


class HalfSpaceTrees(base.AnomalyDetector):
    """Half-Space Trees (HST).

    Half-space trees are an online variant of isolation forests. They work well when anomalies are
    spread out, such as in the `datasets.CreditCard` dataset. However, they do not work well if
    anomalies are packed together in windows, such as in the `datasets.HTTP` and `datasets.SMTP`
    datasets.

    By default, this implementation assumes that each feature has values that are comprised
    between 0 and 1. If this isn't the case, then you can manually specify the limits via the
    `limits` argument. If you do not know the limits in advance, then you can use a
    `preprocessing.MinMaxScaler` as an initial preprocessing step.

    The current implementation builds the trees the first time the `fit_one` method is called.
    Therefore, the first `fit_one` call might be slow, whereas subsequent calls will be very fast
    in comparison. In general, the computation time of both `fit_one` and `score_one` scales
    linearly with the number of trees, and exponentially with the height of each tree.

    Note that high scores indicate anomalies, whereas low scores indicate normal observations.

    Parameters:
        n_trees: Number of trees to use.
        height: Height of each tree. Note that a tree of height `h` is made up of `h + 1`
            levels and therefore contains `2 ** (h + 1) - 1` nodes.
        window_size: Number of observations to use for calculating the mass at each node in
            each tree.
        limits: Specifies the range of each feature. By default each feature is assumed to
            be in range `[0, 1]`.
        seed: Random number seed.

    Example:

        >>> from creme import anomaly

        >>> X = [0.5, 0.45, 0.43, 0.44, 0.445, 0.45, 0.0]
        >>> hst = anomaly.HalfSpaceTrees(
        ...     n_trees=5,
        ...     height=3,
        ...     window_size=3,
        ...     seed=42
        ... )

        >>> for x in X[:3]:
        ...     hst = hst.fit_one({'x': x})  # Warming up

        >>> for x in X:
        ...     features = {'x': x}
        ...     hst = hst.fit_one(features)
        ...     print(f'Anomaly score for x={x:.3f}: {hst.score_one(features):.3f}')
        Anomaly score for x=0.500: 0.107
        Anomaly score for x=0.450: 0.071
        Anomaly score for x=0.430: 0.107
        Anomaly score for x=0.440: 0.107
        Anomaly score for x=0.445: 0.107
        Anomaly score for x=0.450: 0.071
        Anomaly score for x=0.000: 0.853

        The feature values are all comprised between 0 and 1. This is what is assumed by the model
        by default. In the following example, we construct a pipeline that scales the data online
        and ensures that the values of each feature are comprised between 0 and 1.

        >>> from creme import compose
        >>> from creme import datasets
        >>> from creme import metrics
        >>> from creme import preprocessing

        >>> model = compose.Pipeline(
        ...     preprocessing.MinMaxScaler(),
        ...     anomaly.HalfSpaceTrees(seed=42)
        ... )

        >>> auc = metrics.ROCAUC()

        >>> for x, y in datasets.CreditCard().take(8000):
        ...     score = model.score_one(x)
        ...     model = model.fit_one(x, y)
        ...     auc = auc.update(y, score)

        >>> auc
        ROCAUC: 0.940431

    References:
        1. [Tan, S.C., Ting, K.M. and Liu, T.F., 2011, June. Fast anomaly detection for streaming data. In Twenty-Second International Joint Conference on Artificial Intelligence.](https://www.ijcai.org/Proceedings/11/Papers/254.pdf)

    """

    def __init__(self, n_trees=10, height=8, window_size=250,
                 limits: typing.Dict[base.typing.FeatureName, typing.Tuple[float, float]] = None,
                 seed: int = None):

        self.n_trees = n_trees
        self.window_size = window_size
        self.height = height
        self.limits = collections.defaultdict(functools.partial(tuple, (0, 1)))
        if limits is not None:
            self.limits.update(limits)
        self.seed = seed
        self.rng = random.Random(seed)

        self.trees = []
        self.counter = 0
        self._first_window = True

    @property
    def size_limit(self):
        """This is the threshold under which the node search stops during the scoring phase.

        The value .1 is a magic constant indicated in the original paper.

        """
        return .1 * self.window_size

    @property
    def _max_score(self):
        """The largest potential anomaly score."""
        return self.n_trees * self.window_size * (2 ** (self.height + 1) - 1)

    def fit_one(self, x):

        # The trees are built when the first observation comes in
        if not self.trees:
            self.trees = [
                make_padded_tree(
                    limits={i: self.limits[i] for i in x},
                    height=self.height,
                    padding=.15,
                    rng=self.rng,
                    # kwargs
                    r_mass=0,
                    l_mass=0
                )
                for _ in range(self.n_trees)
            ]

        # Update each tree
        for tree in self.trees:
            for node in tree.path(x):
                node.l_mass += 1

        # Pivot the masses if necessary
        self.counter += 1
        if self.counter == self.window_size:
            for tree in self.trees:
                for node, _ in tree.iter_dfs():
                    node.r_mass = node.l_mass
                    node.l_mass = 0
            self._first_window = False
            self.counter = 0

        return self

    def score_one(self, x):

        if self._first_window:
            return 0

        score = 0.
        for tree in self.trees:
            for depth, node in enumerate(tree.path(x)):
                score += node.r_mass * 2 ** depth
                if node.r_mass < self.size_limit:
                    break

        # Normalize the score between 0 and 1
        score /= self._max_score

        # We want high score -> anomaly, but we have high score -> normal
        return 1 - score



from creme import stats


class Link(stats.Univariate):
    """A link joins two univariate statistics as a sequence.

    This can be used to pipe the output of one statistic to the input of another. This can be used,
    for instance, to calculate the mean of the variance of a variable. It can also be used to
    compute shifted statistics by piping statistics with an instance of `stats.Shift`.

    Note that a link is not meant to be instantiated via this class definition. Instead, users can
    link statistics together via the `|` operator.

    Parameters:
        left
        right: The output from `left`'s `get` method is passed to `right`'s `update` method if
            `left`'s `get` method doesn't produce `None.`

    Example:

        >>> from creme import stats
        >>> stat = stats.Shift(1) | stats.Mean()

        No values have been seen, therefore `get` defaults to the initial value of `stats.Mean`,
        which is 0.

        >>> stat.get()
        0.

        Let us now call `update`.

        >>> stat = stat.update(1)

        The output from `get` will still be 0. The reason is that `stats.Shift` has not enough
        values, and therefore outputs it's default value, which is `None`. The `stats.Mean`
        instance is therefore not updated.

        >>> stat.get()
        0.0

        On the next call to `update`, the `stats.Shift` instance has seen enough values, and
        therefore the mean can be updated. The mean is therefore equal to 1, because that's the
        only value from the past.

        >>> stat = stat.update(3)
        >>> stat.get()
        1.0

        On the subsequent call to update, the mean will be updated with the value 3.

        >>> stat = stat.update(4)
        >>> stat.get()
        2.0

        Note that composing statistics returns a new statistic with it's own name.

        >>> stat.name
        'mean_of_shift_1'

    """

    def __init__(self, left: stats.Univariate, right: stats.Univariate):
        self.left = left
        self.right = right

    def update(self, x):
        self.left.update(x)
        y = self.left.get()
        if y is not None:
            self.right.update(y)
        return self

    def get(self):
        return self.right.get()

    @property
    def name(self):
        return f'{self.right.name}_of_{self.left.name}'

    def __repr__(self):
        return repr(self.right)



from scipy.io.arff import arffread

from creme import base

from . import utils


def iter_arff(filepath_or_buffer, target: str = None, compression='infer') -> base.typing.Stream:
    """Iterates over rows from an ARFF file.

    Parameters:
        filepath_or_buffer: Either a string indicating the location of a CSV file, or a buffer
            object that has a `read` method.
        target: Name of the target field.
        compression: For on-the-fly decompression of on-disk data. If this is set to 'infer' and
            `filepath_or_buffer` is a path, then the decompression method is inferred for the
            following extensions: '.gz', '.zip'.

    """

    # If a file is not opened, then we open it
    buffer = filepath_or_buffer
    if not hasattr(buffer, 'read'):
        buffer = utils.open_filepath(buffer, compression)

    try:
        rel, attrs = arffread.read_header(buffer)
    except ValueError as e:
        msg = f'Error while parsing header, error was: {e}'
        raise arffread.ParseArffError(msg)

    names = [attr.name for attr in attrs]
    types = [float if isinstance(attr, arffread.NumericAttribute) else None for attr in attrs]

    for r in buffer:
        x =  {
            name: typ(val) if typ else val
            for name, typ, val in zip(names, types, r.rstrip().split(','))
        }
        try:
            y = x.pop(target) if target else None
        except KeyError as e:
            print(r)
            raise e

        yield x, y

    # Close the file if we opened it
    if buffer is not filepath_or_buffer:
        buffer.close()



"""Anomaly detection.

The estimators in the `anomaly` module have a slightly different API. Instead of a `predict_one`
method, each anomaly detector has a `score_one`. The latter returns an anomaly score for a given
set of features. High scores indicate anomalies, whereas low scores indicate normal observations.
Note that the range of the scores is relative to each estimator.

"""
from .hst import HalfSpaceTrees


__all__ = ['HalfSpaceTrees']



import collections

from .. import stats

from . import base
from . import precision


__all__ = [
    'MacroRecall',
    'MicroRecall',
    'Recall',
    'WeightedRecall'
]


class BaseRecall:

    @property
    def bigger_is_better(self):
        return True

    @property
    def requires_labels(self):
        return True


class Recall(BaseRecall, base.BinaryMetric):
    """Binary recall score.

    Example:

        >>> from creme import metrics

        >>> y_true = [True, False, True, True, True]
        >>> y_pred = [True, True, False, True, True]

        >>> metric = metrics.Recall()

        >>> for yt, yp in zip(y_true, y_pred):
        ...     print(metric.update(yt, yp))
        Recall: 1.
        Recall: 1.
        Recall: 0.5
        Recall: 0.666667
        Recall: 0.75

    """

    def __init__(self):
        self._mean = stats.Mean()

    def update(self, y_true, y_pred, sample_weight=1.):
        if y_true:
            self._mean.update(x=y_true == y_pred, w=sample_weight)
        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        if y_true:
            self._mean.revert(x=y_true == y_pred, w=sample_weight)
        return self

    def get(self):
        return self._mean.get()


class MacroRecall(BaseRecall, base.MultiClassMetric):
    """Macro-average recall score.

    Example:

        >>> from creme import metrics

        >>> y_true = [0, 1, 2, 2, 2]
        >>> y_pred = [0, 0, 2, 2, 1]

        >>> metric = metrics.MacroRecall()

        >>> for yt, yp in zip(y_true, y_pred):
        ...     print(metric.update(yt, yp))
        MacroRecall: 1.
        MacroRecall: 0.5
        MacroRecall: 0.666667
        MacroRecall: 0.666667
        MacroRecall: 0.555556

    """

    def __init__(self):
        self.recalls = collections.defaultdict(Recall)
        self._class_counts = collections.Counter()

    def update(self, y_true, y_pred, sample_weight=1.):
        self.recalls[y_true].update(True, y_true == y_pred, sample_weight)
        self._class_counts.update([y_true, y_pred])
        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        self.recalls[y_true].revert(True, y_true == y_pred, sample_weight)
        self._class_counts.subtract([y_true, y_pred])
        return self

    def get(self):
        relevant = [c for c, count in self._class_counts.items() if count > 0]
        try:
            return sum(self.recalls[c].get() for c in relevant) / len(relevant)
        except ZeroDivisionError:
            return 0.


class MicroRecall(precision.MicroPrecision):
    """Micro-average recall score.

    The micro-average recall is exactly equivalent to the micro-average precision as well as the
    micro-average F1 score.

    Example:

        >>> from creme import metrics

        >>> y_true = [0, 1, 2, 2, 2]
        >>> y_pred = [0, 0, 2, 2, 1]

        >>> metric = metrics.MicroRecall()

        >>> for yt, yp in zip(y_true, y_pred):
        ...     print(metric.update(yt, yp))
        MicroRecall: 1.
        MicroRecall: 0.5
        MicroRecall: 0.666667
        MicroRecall: 0.75
        MicroRecall: 0.6

    References:
        1. [Why are precision, recall and F1 score equal when using micro averaging in a multi-class problem?](https://simonhessner.de/why-are-precision-recall-and-f1-score-equal-when-using-micro-averaging-in-a-multi-class-problem/)

    """


class WeightedRecall(BaseRecall, base.MultiClassMetric):
    """Weighted-average recall score.

    This uses the support of each label to compute an average score, whereas `MacroRecall`
    ignores the support.

    Example:

        >>> from creme import metrics

        >>> y_true = [0, 1, 2, 2, 2]
        >>> y_pred = [0, 0, 2, 2, 1]

        >>> metric = metrics.WeightedRecall()

        >>> for yt, yp in zip(y_true, y_pred):
        ...     print(metric.update(yt, yp))
        WeightedRecall: 1.
        WeightedRecall: 0.5
        WeightedRecall: 0.666667
        WeightedRecall: 0.75
        WeightedRecall: 0.6

    """

    def __init__(self):
        self.recalls = collections.defaultdict(Recall)
        self.support = collections.Counter()
        self._class_counts = collections.Counter()

    def update(self, y_true, y_pred, sample_weight=1.):
        self.recalls[y_true].update(True, y_true == y_pred, sample_weight)
        self.support.update({y_true: sample_weight})
        self._class_counts.update([y_true, y_pred])
        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        self.recalls[y_true].revert(True, y_true == y_pred, sample_weight)
        self.support.subtract({y_true: sample_weight})
        self._class_counts.subtract([y_true, y_pred])
        return self

    def get(self):
        relevant = [c for c, count in self._class_counts.items() if count > 0]
        try:
            return (
                sum(self.recalls[c].get() * self.support[c] for c in relevant) /
                sum(self.support[c] for c in relevant)
            )
        except ZeroDivisionError:
            return 0.



"""Feature selection."""
from .k_best import SelectKBest
from .random import PoissonInclusion
from .variance import VarianceThreshold


__all__ = [
    'PoissonInclusion',
    'SelectKBest',
    'VarianceThreshold'
]



import collections

from creme import base
from creme import stats


class VarianceThreshold(base.Transformer):
    """Removes low-variance features.

    Parameters:
        threshold: Only features with a variance above the threshold will be kept.
        min_samples: The minimum number of samples required to perform selection.

    Attributes:
        variances (dict): The variance of each feature.

    Example:

        >>> from creme import feature_selection
        >>> from creme import stream

        >>> X = [
        ...     [0, 2, 0, 3],
        ...     [0, 1, 4, 3],
        ...     [0, 1, 1, 3]
        ... ]

        >>> selector = feature_selection.VarianceThreshold()

        >>> for x, _ in stream.iter_array(X):
        ...     print(selector.fit_one(x).transform_one(x))
        {0: 0, 1: 2, 2: 0, 3: 3}
        {1: 1, 2: 4}
        {1: 1, 2: 1}

    """

    def __init__(self, threshold=0, min_samples=2):
        self.threshold = threshold
        self.min_samples = min_samples
        self.variances = collections.defaultdict(stats.Var)

    def fit_one(self, x):

        for i, xi in x.items():
            self.variances[i].update(xi)

        return self

    def check_feature(self, feature):
        if feature not in self.variances:
            return True
        if self.variances[feature].mean.n < self.min_samples:
            return True
        return self.variances[feature].get() > self.threshold

    def transform_one(self, x):
        return {
            i: xi
            for i, xi in x.items()
            if self.check_feature(i)
        }



"""
The tests performed here confirm that the outputs of the online LDA are exactly the same as those of
the original with a batch_size of size 1. Coverage is 100%.

References:
    1. Jordan Boyd-Graber, Ke Zhai, Online Latent Dirichlet Allocation with Infinite Vocabulary.
    http://proceedings.mlr.press/v28/zhai13.pdf

    2. Creme's Online LDA reproduces exactly the same results of the original one with a size of
    batch 1:
    https://github.com/kzhai/PyInfVoc.
"""
import numpy as np

from creme.decomposition import LDA


DOC_SET = [
    'weather cold',
    'weather hot dry',
    'weather cold rainny',
    'weather hot',
    'weather cold humid',
]

REFERENCE_STATISTICS_TWO_COMPONENTS = [
    {
        0: np.array([0., 0., 0.]),
        1: np.array([0., 1., 1.]),
    },
    {
        0: np.array([0., 0.6, 0., 0.8, 0.6]),
        1: np.array([0., 0.4, 0., 0.2, 0.4])},
    {
        0: np.array([0., 0., 0., 0., 0., 0.]),
        1: np.array([0., 1., 1., 0., 0., 1.])},
    {
        0: np.array([0., 0.2, 0., 0.6, 0., 0.]),
        1: np.array([0., 0.8, 0., 0.4, 0., 0.])},
    {
        0: np.array([0., 0., 0.2, 0., 0., 0., 0.2]),
        1: np.array([0., 1., 0.8, 0., 0., 0., 0.8]),
    },
]


REFERENCE_STATISTICS_FIVE_COMPONENTS = [
    {
        0: np.array([0., 0.4, 0.2]),
        1: np.array([0., 0.2, 0.6]),
        2: np.array([0., 0.4, 0.]),
        3: np.array([0., 0., 0.]),
        4: np.array([0., 0., 0.2])
    },
    {
        0: np.array([0., 0.8, 0., 0.4, 0.6]),
        1: np.array([0., 0., 0., 0.2, 0.4]),
        2: np.array([0., 0., 0., 0., 0.]),
        3: np.array([0., 0., 0., 0.2, 0.]),
        4: np.array([0., 0.2, 0., 0.2, 0.])
    },
    {
        0: np.array([0., 0.4, 0.2, 0., 0., 0.]),
        1: np.array([0., 0.2, 0.6, 0., 0., 0.6]),
        2: np.array([0., 0.4, 0.2, 0., 0., 0.4]),
        3: np.array([0., 0., 0., 0., 0., 0.]),
        4: np.array([0., 0., 0., 0., 0., 0.])
    },
    {
        0: np.array([0., 0.2, 0., 0.4, 0., 0.]),
        1: np.array([0., 0.2, 0., 0.2, 0., 0.]),
        2: np.array([0., 0.4, 0., 0.2, 0., 0.]),
        3: np.array([0., 0.2, 0., 0.2, 0., 0.]),
        4: np.array([0., 0., 0., 0., 0., 0.])
    },
    {
        0: np.array([0., 0., 0.2, 0., 0., 0., 0.2]),
        1: np.array([0., 0., 0., 0., 0., 0., 0.]),
        2: np.array([0., 0.8, 0.8, 0., 0., 0., 0.6]),
        3: np.array([0., 0.2, 0., 0., 0., 0., 0.2]),
        4: np.array([0., 0., 0., 0., 0., 0., 0.])
    }
]

REFERENCE_FIVE_COMPONENTS = [
    np.array([1.5, 0.5, 0.5, 0.5, 1.5]),
    np.array([1.5, 1.5, 0.5, 0.5, 1.5]),
    np.array([0.5, 0.5, 3.5, 0.5, 0.5]),
    np.array([0.5, 0.5, 0.5, 2.5, 0.5]),
    np.array([2.5, 0.5, 0.5, 1.5, 0.5]),

]

REFERENCE_COMPONENTS_WITH_PRUNNING = [
    np.array([0.5, 2.5]),
    np.array([3.5, 0.5]),
    np.array([0.5, 3.5]),
    np.array([2.5, 0.5]),
    np.array([2.5, 1.5]),
]

REFERENCE_FIT_ONE_PREDICT_ONE = [
    np.array([2.5, 0.5]),
    np.array([2.5, 1.5]),
    np.array([1.5, 2.5]),
    np.array([0.5, 2.5]),
    np.array([0.5, 3.5]),
]


def test_extraction_words_ids():
    '''
    Assert that inputs words are splitted.
    Assert that indexes are updated and extractable.
    '''
    np.random.seed(42)

    lda = LDA(2, number_of_documents=5)

    word_indexes_list = []

    for doc in DOC_SET:

        words = lda.process_text(doc)

        lda._update_indexes(word_list=words)

        word_indexes_list.append(
            [lda.word_to_index[word] for word in words]
        )

    assert word_indexes_list == [
        [1, 2],
        [1, 3, 4],
        [1, 2, 5],
        [1, 3],
        [1, 2, 6],
    ]


def test_statistics_two_components():
    '''
    Assert that online lda extracts waited statistics on current document.
    '''
    n_components = 2

    np.random.seed(42)

    lda = LDA(n_components, number_of_documents=60)

    statistics_list = []

    for doc in DOC_SET:

        word_list = lda.process_text(doc)

        lda._update_indexes(word_list=word_list)

        word_indexes = [lda.word_to_index[word] for word in word_list]

        statistics, _ = lda._compute_statistics_components(
            words_indexes_list=word_indexes,
        )

        statistics_list.append(statistics)

        lda._update_weights(
            statistics=statistics
        )

    for index, statistics in enumerate(statistics_list):
        for component in range(n_components):
            assert np.array_equal(
                a1=statistics[component],
                a2=REFERENCE_STATISTICS_TWO_COMPONENTS[index][component],
            )


def test_statistics_five_components():
    '''
    Assert that online lda extracts waited statistics on current document.
    '''
    np.random.seed(42)

    n_components = 5

    lda = LDA(
        n_components=n_components,
        number_of_documents=60,
        maximum_size_vocabulary=100,
        alpha_beta=100,
        alpha_theta=0.5,
    )

    statistics_list = []

    for doc in DOC_SET:

        word_list = lda.process_text(doc)

        lda._update_indexes(word_list=word_list)

        word_indexes = [lda.word_to_index[word] for word in word_list]

        statistics, _ = lda._compute_statistics_components(
            words_indexes_list=word_indexes,
        )

        statistics_list.append(statistics)

        lda._update_weights(
            statistics=statistics
        )

    for index, statistics in enumerate(statistics_list):
        for component in range(n_components):
            assert np.array_equal(
                a1=statistics[component],
                a2=REFERENCE_STATISTICS_FIVE_COMPONENTS[index][component],
            )


def test_five_components():
    '''
    Assert that components computed are identical to the original version for n dimensions.
    '''
    np.random.seed(42)

    n_components = 5

    online_lda = LDA(
        n_components=n_components,
        number_of_documents=60,
        maximum_size_vocabulary=100,
        alpha_beta=100,
        alpha_theta=0.5,
    )

    components_list = []

    for document in DOC_SET:
        components_list.append(online_lda.fit_transform_one(document))

    for index, component in enumerate(components_list):
        assert np.array_equal(
            a1=list(component.values()),
            a2=REFERENCE_FIVE_COMPONENTS[index],
        )


def test_prunning_vocabulary():
    '''
    Vocabulary prunning is available to improve accuracy and limit memory usage.
    You can perform vocabulary prunning with parameters vocab_prune_interval (int) and
    maximum_size_vocabulary (int).
    '''
    np.random.seed(42)

    online_lda = LDA(
        n_components=2,
        number_of_documents=60,
        vocab_prune_interval=2,
        maximum_size_vocabulary=3
    )

    components_list = []

    for document in DOC_SET:
        components_list.append(
            online_lda.fit_transform_one(x=document)
        )

    for index, component in enumerate(components_list):
        assert np.array_equal(
            a1=list(component.values()),
            a2=REFERENCE_COMPONENTS_WITH_PRUNNING[index]
        )


def test_fit_transform():
    '''
    Assert that fit_one and transform_one methods returns waited ouput.
    '''
    np.random.seed(42)

    online_lda = LDA(
        n_components=2,
        number_of_documents=60,
        vocab_prune_interval=2,
        maximum_size_vocabulary=3,
    )
    components_list = []

    for document in DOC_SET:
        online_lda = online_lda.fit_one(x=document)

        components_list.append(
            online_lda.transform_one(x=document)
        )

    for index, component in enumerate(components_list):
        assert np.array_equal(
            a1=list(component.values()),
            a2=REFERENCE_FIT_ONE_PREDICT_ONE[index]
        )



import collections
import copy
import functools
import typing

import numpy as np

from creme import stats
from creme import optim
from creme import utils

from . import base


__all__ = ['BiasedMF']


class BiasedMF(base.Recommender):
    """Biased Matrix Factorization for recommender systems.

    The model equation is defined as:

    $$\\hat{y}(x) = \\bar{y} + bu_{u} + bi_{i} + \\langle \\mathbf{v}_u, \\mathbf{v}_i \\rangle$$

    Where :math:`bu_{u}` and :math:`bi_{i}` are respectively the user and item biases. The last term
    being simply the dot product between the latent vectors of the given user-item pair:

    $$\\langle \\mathbf{v}_u, \\mathbf{v}_i \\rangle = \\sum_{f=1}^{k} \\mathbf{v}_{u, f} \\cdot \\mathbf{v}_{i, f}$$

    Where :math:`k` is the number of latent factors. The model expect dict inputs containing both a
    `user` and an `item` entries.

    Parameters:
        n_factors: Dimensionality of the factorization or number of latent factors.
        bias_optimizer: The sequential optimizer used for updating the bias
            weights.
        latent_optimizer: The sequential optimizer used for updating the latent weights.
        loss: The loss function to optimize for.
        l2_bias: Amount of L2 regularization used to push bias weights towards 0.
        l2_latent: Amount of L2 regularization used to push latent weights towards 0.
        weight_initializer: Weights initialization scheme.
        latent_initializer: Latent factors initialization scheme.
        clip_gradient: Clips the absolute value of each gradient value.
        seed: Randomization seed used for reproducibility.

    Attributes:
        global_mean (stats.Mean): The target arithmetic mean.
        u_biases (collections.defaultdict): The user bias weights.
        i_biases (collections.defaultdict): The item bias weights.
        u_latents (collections.defaultdict): The user latent vectors randomly initialized.
        i_latents (collections.defaultdict): The item latent vectors randomly initialized.
        u_bias_optimizer (optim.Optimizer): The sequential optimizer used for updating the user bias
            weights.
        i_bias_optimizer (optim.Optimizer): The sequential optimizer used for updating the item bias
            weights.
        u_latent_optimizer (optim.Optimizer): The sequential optimizer used for updating the user
            latent weights.
        i_latent_optimizer (optim.Optimizer): The sequential optimizer used for updating the item
            latent weights.

    Example:

        >>> from creme import optim
        >>> from creme import reco

        >>> X_y = (
        ...     ({'user': 'Alice', 'item': 'Superman'}, 8),
        ...     ({'user': 'Alice', 'item': 'Terminator'}, 9),
        ...     ({'user': 'Alice', 'item': 'Star Wars'}, 8),
        ...     ({'user': 'Alice', 'item': 'Notting Hill'}, 2),
        ...     ({'user': 'Alice', 'item': 'Harry Potter'}, 5),
        ...     ({'user': 'Bob', 'item': 'Superman'}, 8),
        ...     ({'user': 'Bob', 'item': 'Terminator'}, 9),
        ...     ({'user': 'Bob', 'item': 'Star Wars'}, 8),
        ...     ({'user': 'Bob', 'item': 'Notting Hill'}, 2)
        ... )

        >>> model = reco.BiasedMF(
        ...     n_factors=10,
        ...     bias_optimizer=optim.SGD(0.025),
        ...     latent_optimizer=optim.SGD(0.025),
        ...     latent_initializer=optim.initializers.Normal(mu=0., sigma=0.1, seed=71)
        ... )

        >>> for x, y in X_y:
        ...     _ = model.fit_one(x, y)

        >>> model.predict_one({'user': 'Bob', 'item': 'Harry Potter'})
        6.489025

    .. note::
        This model expects a dict input with a `user` and an `item` entries without any type
        constraint on their values (i.e. can be strings or numbers). Other entries are ignored.

    References:
        1. [Paterek, A., 2007, August. Improving regularized singular value decomposition for collaborative filtering. In Proceedings of KDD cup and workshop (Vol. 2007, pp. 5-8)](https://www.cs.uic.edu/~liub/KDD-cup-2007/proceedings/Regular-Paterek.pdf)
        2. [Matrix factorization techniques for recommender systems](https://datajobs.com/data-science-repo/Recommender-Systems-[Netflix].pdf)

    """

    def __init__(self, n_factors=10, bias_optimizer: optim.Optimizer = None,
                 latent_optimizer: optim.Optimizer = None, loss: optim.losses.Loss = None,
                 l2_bias=0., l2_latent=0.,
                 weight_initializer: optim.initializers.Initializer = None,
                 latent_initializer: optim.initializers.Initializer = None,
                 clip_gradient=1e12, seed: int = None):

        self.n_factors = n_factors
        self.u_bias_optimizer = optim.SGD() if bias_optimizer is None else copy.deepcopy(bias_optimizer)
        self.i_bias_optimizer = optim.SGD() if bias_optimizer is None else copy.deepcopy(bias_optimizer)
        self.u_latent_optimizer = optim.SGD() if latent_optimizer is None else copy.deepcopy(latent_optimizer)
        self.i_latent_optimizer = optim.SGD() if latent_optimizer is None else copy.deepcopy(latent_optimizer)
        self.loss = optim.losses.Squared() if loss is None else loss
        self.l2_bias = l2_bias
        self.l2_latent = l2_latent

        if weight_initializer is None:
            weight_initializer = optim.initializers.Zeros()
        self.weight_initializer = weight_initializer

        if latent_initializer is None:
            latent_initializer = optim.initializers.Normal(sigma=.1, seed=seed)
        self.latent_initializer = latent_initializer

        self.clip_gradient = clip_gradient
        self.seed = seed
        self.global_mean = stats.Mean()

        self.u_biases: typing.DefaultDict[int, optim.initializers.Initializer] = collections.defaultdict(weight_initializer)
        self.i_biases: typing.DefaultDict[int, optim.initializers.Initializer] = collections.defaultdict(weight_initializer)

        random_latents = functools.partial(
            self.latent_initializer,
            shape=self.n_factors
        )
        self.u_latents: typing.DefaultDict[int, optim.initializers.Initializer] = collections.defaultdict(random_latents)
        self.i_latents: typing.DefaultDict[int, optim.initializers.Initializer] = collections.defaultdict(random_latents)

    def _predict_one(self, user, item):

        # Initialize the prediction to the mean
        y_pred = self.global_mean.get()

        # Add the user bias
        y_pred += self.u_biases[user]

        # Add the item bias
        y_pred += self.i_biases[item]

        # Add the dot product of the user and the item latent vectors
        y_pred += np.dot(self.u_latents[user], self.i_latents[item])

        return y_pred

    def _fit_one(self, user, item, y):

        # Update the global mean
        self.global_mean.update(y)

        # Calculate the gradient of the loss with respect to the prediction
        g_loss = self.loss.gradient(y, self._predict_one(user, item))

        # Clamp the gradient to avoid numerical instability
        g_loss = utils.math.clamp(g_loss, minimum=-self.clip_gradient, maximum=self.clip_gradient)

        # Calculate weights gradients
        u_grad_bias = {user: g_loss + self.l2_bias * self.u_biases[user]}
        i_grad_bias = {item: g_loss + self.l2_bias * self.i_biases[item]}
        u_latent_grad = {user: g_loss * self.i_latents[item] + self.l2_latent * self.u_latents[user]}
        i_latent_grad = {item: g_loss * self.u_latents[user] + self.l2_latent * self.i_latents[item]}

        # Update weights
        self.u_biases = self.u_bias_optimizer.update_after_pred(self.u_biases, u_grad_bias)
        self.i_biases = self.i_bias_optimizer.update_after_pred(self.i_biases, i_grad_bias)
        self.u_latents = self.u_latent_optimizer.update_after_pred(self.u_latents, u_latent_grad)
        self.i_latents = self.i_latent_optimizer.update_after_pred(self.i_latents, i_latent_grad)

        return self



"""Feature extraction."""
from .agg import Agg
from .agg import TargetAgg
from .vectorize import BagOfWords
from .vectorize import TFIDF


__all__ = [
    'Agg',
    'BagOfWords',
    'TargetAgg',
    'TFIDF'
]



import collections
import math

from . import base


__all__ = ['BernoulliNB']


class BernoulliNB(base.BaseNB):
    """Bernoulli Naive Bayes.

    Parameters:
        alpha: Additive (Laplace/Lidstone) smoothing parameter (use 0 for no smoothing).
        true_threshold: Threshold for binarizing (mapping to booleans) features.

    Attributes:
        class_counts (collections.Counter): Number of times each class has been seen.
        feature_counts (collections.defaultdict): Total frequencies per feature and class.

    Example:

        >>> import math
        >>> from creme import compose
        >>> from creme import feature_extraction
        >>> from creme import naive_bayes

        >>> docs = [
        ...     ('Chinese Beijing Chinese', 'yes'),
        ...     ('Chinese Chinese Shanghai', 'yes'),
        ...     ('Chinese Macao', 'yes'),
        ...     ('Tokyo Japan Chinese', 'no')
        ... ]
        >>> model = compose.Pipeline(
        ...     ('tokenize', feature_extraction.BagOfWords(lowercase=False)),
        ...     ('nb', naive_bayes.BernoulliNB(alpha=1))
        ... )
        >>> for sentence, label in docs:
        ...     model = model.fit_one(sentence, label)

        >>> model['nb'].p_class('yes')
        0.75
        >>> model['nb'].p_class('no')
        0.25

        >>> cp = model['nb'].p_feature_given_class

        >>> cp('Chinese', 'yes') == (3 + 1) / (3 + 2)
        True

        >>> cp('Japan', 'yes') == (0 + 1) / (3 + 2)
        True
        >>> cp('Tokyo', 'yes') == (0 + 1) / (3 + 2)
        True

        >>> cp('Beijing', 'yes') == (1 + 1) / (3 + 2)
        True
        >>> cp('Macao', 'yes') == (1 + 1) / (3 + 2)
        True
        >>> cp('Shanghai', 'yes') == (1 + 1) / (3 + 2)
        True

        >>> cp('Chinese', 'no') == (1 + 1) / (1 + 2)
        True

        >>> cp('Japan', 'no') == (1 + 1) / (1 + 2)
        True
        >>> cp('Tokyo', 'no') == (1 + 1) / (1 + 2)
        True

        >>> cp('Beijing', 'no') == (0 + 1) / (1 + 2)
        True
        >>> cp('Macao', 'no') == (0 + 1) / (1 + 2)
        True
        >>> cp('Shanghai', 'no') == (0 + 1) / (1 + 2)
        True

        >>> new_text = 'Chinese Chinese Chinese Tokyo Japan'
        >>> tokens = model['tokenize'].transform_one(new_text)
        >>> jlh = model['nb'].joint_log_likelihood(tokens)
        >>> math.exp(jlh['yes'])
        0.005184
        >>> math.exp(jlh['no'])
        0.021947
        >>> model.predict_one(new_text)
        'no'

    References:
        1. [The Bernoulli model](https://nlp.stanford.edu/IR-book/html/htmledition/the-bernoulli-model-1.html)

    """

    def __init__(self, alpha=1., true_threshold=0.):
        self.alpha = alpha
        self.true_threshold = true_threshold
        self.class_counts = collections.Counter()
        self.feature_counts = collections.defaultdict(collections.Counter)

    def fit_one(self, x, y):
        self.class_counts.update((y,))

        for i, xi in x.items():
            self.feature_counts[i].update({y: xi > self.true_threshold})

        return self

    def p_feature_given_class(self, f: str, c: str) -> float:
        num = self.feature_counts.get(f, {}).get(c, 0.) + self.alpha
        den = self.class_counts[c] + self.alpha * 2
        return num / den

    def p_class(self, c: str) -> float:
        return self.class_counts[c] / sum(self.class_counts.values())

    def joint_log_likelihood(self, x):
        return {
            c: math.log(self.p_class(c)) + sum(map(
                math.log,
                (
                    10e-10 + self.p_feature_given_class(f, c)
                    if f in x and x[f] > self.true_threshold
                    else 10e-10 + (1. - self.p_feature_given_class(f, c))
                    for f in self.feature_counts
                )
            ))
            for c in self.class_counts
        }



import random

from creme import base


__all__ = ['PoissonInclusion']


class PoissonInclusion(base.Transformer):
    """Randomly selects features with an inclusion trial.

    When a new feature is encoutered, it is selected with probability `p`. The number of times a
    feature needs to beseen before it is added to the model follows a geometric distribution with
    expected value `1 / p`. This feature selection method is meant to be used when you have a
    very large amount of sparse features.

    Parameters:
        p: Probability of including a feature the first time it is encoutered.

    References:
        1. [McMahan, H.B., Holt, G., Sculley, D., Young, M., Ebner, D., Grady, J., Nie, L., Phillips, T., Davydov, E., Golovin, D. and Chikkerur, S., 2013, August. Ad click prediction: a view from the trenches. In Proceedings of the 19th ACM SIGKDD international conference on Knowledge discovery and data mining (pp. 1222-1230)](https://static.googleusercontent.com/media/research.google.com/en//pubs/archive/41159.pdf)

    """

    def __init__(self, p: float, seed=None):
        self.p = p
        self.seed = seed
        self.rng = random.Random(seed)
        self.included = set()

    def transform_one(self, x):

        xt = {}

        for i, xi in x.items():
            if i not in self.included and self.rng.random() < self.p:
                self.included.add(i)
                xt[i] = xi
            xt[i] = xi

        return xt



from . import base


__all__ = ['SGD']


class SGD(base.Optimizer):
    """Plain stochastic gradient descent.

    Parameters:
        lr

    Example:

        >>> from creme import datasets
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import optim
        >>> from creme import preprocessing

        >>> X_y = datasets.Phishing()
        >>> optimizer = optim.SGD(0.1)
        >>> model = (
        ...     preprocessing.StandardScaler() |
        ...     linear_model.LogisticRegression(optimizer)
        ... )
        >>> metric = metrics.F1()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        F1: 0.878521

    References:
        1. [Robbins, H. and Monro, S., 1951. A stochastic approximation method. The annals of mathematical statistics, pp.400-407](https://pdfs.semanticscholar.org/34dd/d8865569c2c32dec9bf7ffc817ff42faaa01.pdf)

    """

    def __init__(self, lr=.01):
        super().__init__(lr)

    def _update_after_pred(self, w, g):

        for i, gi in g.items():
            w[i] -= self.learning_rate * gi

        return w



"""Generic branch and leaf implementation."""
import collections
import operator
import textwrap
import typing


class Op:
    """An operator that is part of a split."""

    __slots__ = 'symbol', 'func'

    def __init__(self, symbol, func):
        self.symbol = symbol
        self.func = func

    def __call__(self, a, b):
        return self.func(a, b)

    def __repr__(self):
        return self.symbol


LT = Op('<', operator.lt)
EQ = Op('=', operator.eq)


class Split:
    """A split that is part of a branch."""

    def __init__(self, on, how, at):
        self.on = on
        self.how = how
        self.at = at

    def __call__(self, x):
        return self.how(x[self.on], self.at)

    def __repr__(self):
        at = self.at
        if isinstance(at, float):
            at = f'{at:.5f}'
        return f'{self.on} {repr(self.how)} {at}'


class Node:
    """A node is a branch or a leaf."""

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class Branch(Node):
    """A branch is made up of a split and two child nodes."""

    def __init__(self, split, left, right, **kwargs):
        super().__init__(**kwargs)
        self.split = split
        self.left = left
        self.right = right

    def next(self, x):
        """Returns the next node where x belongs."""
        return self.left if self.split(x) else self.right

    def path(self, x):
        """Iterate over the nodes that lead to the leaf which contains x."""
        yield self
        yield from self.next(x).path(x)

    def __repr__(self):
        """

        Example:

            >>> tree = Branch(
            ...     Split('x', LT, 17.42),
            ...     Branch(
            ...         Split('y', LT, -4.38),
            ...         Leaf(no=2),
            ...         Leaf(no=3),
            ...         no=1
            ...     ),
            ...     Leaf(no=4),
            ...     no=0
            ... )

            >>> tree
            x < 17.42000 {'no': 0}
                y < -4.38000 {'no': 1}
                {'no': 2}
                {'no': 3}
                {'no': 4}

        """
        return (
            repr(self.split) + ' ' +
            str({k: v for k, v in self.__dict__.items() if k not in ('split', 'left', 'right')}) +
            textwrap.indent(f'\n{self.left}\n{self.right}', prefix=' ' * 2)
        )

    @property
    def size(self):
        """Number of descendants, including thyself."""
        return 1 + self.left.size + self.right.size

    @property
    def height(self):
        """Distance to the deepest node."""
        return 1 + max(self.left.height, self.right.height)

    def iter_dfs(self, depth=0):
        """Iterate over nodes in a depth-first manner.

        Example:

            >>> tree = Branch(
            ...     None,
            ...     Branch(
            ...         None,
            ...         Leaf(no=2),
            ...         Leaf(no=3),
            ...         no=1
            ...     ),
            ...     Leaf(no=4),
            ...     no=0
            ... )

            >>> for node, depth in tree.iter_dfs():
            ...     print(f'#{node.no}, depth {depth}')
            #0, depth 0
            #1, depth 1
            #2, depth 2
            #3, depth 2
            #4, depth 1

        """
        yield self, depth
        yield from self.left.iter_dfs(depth=depth + 1)
        yield from self.right.iter_dfs(depth=depth + 1)


    def iter_leaves(self):
        """Iterate over the leaves in a depth-first manner.

        Example:

            ::

                >>> tree = Branch(
                ...     None,
                ...     Branch(
                ...         None,
                ...         Leaf(no=2),
                ...         Leaf(no=3),
                ...         no=1
                ...     ),
                ...     Leaf(no=4),
                ...     no=0
                ... )

                >>> for leaf in tree.iter_leaves():
                ...     print(f'#{leaf.no}')
                #2
                #3
                #4

        """
        yield from self.left.iter_leaves()
        yield from self.right.iter_leaves()

    def iter_branches(self):
        """Iterate over branches in a depth-first manner.

        Example:

            ::

                >>> tree = Branch(
                ...     None,
                ...     Branch(
                ...         None,
                ...         Leaf(no=2),
                ...         Leaf(no=3),
                ...         no=1
                ...     ),
                ...     Leaf(no=4),
                ...     no=0
                ... )

                >>> for branch in tree.iter_branches():
                ...     print(f'#{branch.no}')
                #0
                #1

        """
        yield self
        yield from self.left.iter_branches()
        yield from self.right.iter_branches()

    def iter_edges(self):
        """Iterate over edges in a depth-first manner.

        Example:

            ::

                >>> tree = Branch(
                ...     None,
                ...     Branch(
                ...         None,
                ...         Leaf(no=2),
                ...         Leaf(no=3),
                ...         no=1
                ...     ),
                ...     Leaf(no=4),
                ...     no=0
                ... )

                >>> for parent_no, child_no, parent, child, child_depth in tree.iter_edges():
                ...     print(parent_no, child_no, child_depth)
                None 0 0
                0 1 1
                1 2 2
                1 3 2
                0 4 1

        """
        counter = 0

        def iterate(node, depth):

            nonlocal counter
            no = counter

            if isinstance(node, Branch):
                for child in (node.left, node.right):
                    counter += 1
                    yield no, counter, node, child, depth + 1
                    if isinstance(child, Branch):
                        yield from iterate(child, depth=depth + 1)

        yield None, 0, None, self, 0
        yield from iterate(self, depth=0)

    def iter_blocks(self, limits: typing.Dict[typing.Hashable, typing.Tuple[float, float]], depth=-1):
        """Iterate over the blocks which enclose each node.

        This only makes sense if the branches of the provided tree use the ``<`` operator as a
        split rule.

        Parameters:
            limits (dict)
            depth (int): Desired tree depth. Set to ``-1`` to iterate over the leaves.

        Example:

            ::

                >>> import operator
                >>> tree = Branch(
                ...     Split('x', operator.lt, .5),
                ...     Leaf(no=0),
                ...     Branch(
                ...         Split('y', operator.lt, .5),
                ...         Leaf(no=1),
                ...         Leaf(no=2)
                ...     )
                ... )

                >>> for leaf, block in tree.iter_blocks(limits={'x': (0, 1), 'y': (0, 1)}):
                ...     print(leaf.no, block)
                0 {'x': (0, 0.5), 'y': (0, 1)}
                1 {'x': (0.5, 1), 'y': (0, 0.5)}
                2 {'x': (0.5, 1), 'y': (0.5, 1)}

        """
        if depth == 0:
            yield (self, limits)
            return

        on, at = self.split.on, self.split.at

        l_limits = {**limits, on: (limits[on][0], at)} if on in limits else limits
        yield from self.left.iter_blocks(limits=l_limits, depth=depth - 1)

        r_limits = {**limits, on: (at, limits[on][1])} if on in limits else limits
        yield from self.right.iter_blocks(limits=r_limits, depth=depth - 1)

    def iter_splits(self, limits: typing.Dict[typing.Hashable, typing.Tuple[float, float]]):
        """Iterate over splits.

        This only makes sense if the branches of the provided tree use the ``<`` operator as a split
        rule.

        Parameters:
            limits (dict)
            depth (int)

        Example:

            ::

                >>> import operator
                >>> tree = Branch(
                ...     Split('x', operator.lt, .5),
                ...     Leaf(no=0),
                ...     Branch(
                ...         Split('y', operator.lt, .5),
                ...         Leaf(no=1),
                ...         Leaf(no=2)
                ...     )
                ... )

                >>> for line in tree.iter_splits({'x': (0, 1), 'y': (0, 1)}):
                ...     print(line)
                {'x': (0.5, 0.5), 'y': (0, 1)}
                {'x': (0.5, 1), 'y': (0.5, 0.5)}

        """

        on, at = self.split.on, self.split.at

        yield {**limits, on: (at, at)}

        yield from self.left.iter_splits(limits={**limits, on: (limits[on][0], at)})
        yield from self.right.iter_splits(limits={**limits, on: (at, limits[on][1])})


class Leaf(Node):

    def __repr__(self):
        return str(self.__dict__)

    def path(self, x):
        yield self

    @property
    def size(self):
        return 1

    @property
    def height(self):
        return 0

    def iter_dfs(self, depth=0):
        yield self, depth

    def iter_leaves(self):
        yield self

    def iter_branches(self):
        yield from ()

    def iter_edges(self):
        yield None, 0, None, self, 0

    def iter_blocks(self, limits, depth=-1):
        yield self, limits

    def iter_splits(self, limits):
        yield from ()



from . import base

import collections


__all__ = ['ConfusionMatrix']


class ConfusionMatrix(base.MultiClassMetric):
    """Confusion matrix.

    This class is different from the rest of the classes from the `metrics` module in that it
    doesn't have a `get` method.

    Attributes:
        classes: The entire set of seen classes, whether they are part of the predictions or the
            true values.

    Example:

        >>> from creme import metrics

        >>> y_true = ['cat', 'ant', 'cat', 'cat', 'ant', 'bird']
        >>> y_pred = ['ant', 'ant', 'cat', 'cat', 'ant', 'cat']

        >>> cm = metrics.ConfusionMatrix()

        >>> for y_t, y_p in zip(y_true, y_pred):
        ...     cm = cm.update(y_t, y_p)

        >>> cm
                    ant  bird   cat
            ant     2     0     0
            bird     0     0     1
            cat     1     0     2

        >>> cm['bird']['cat']
        1.0

    """

    _fmt = '0.0f'

    @property
    def bigger_is_better(self):
        return None

    @property
    def requires_labels(self):
        return True

    def get(self):
        return self

    def __init__(self):
        self.counts = collections.defaultdict(collections.Counter)
        self.class_counts = collections.Counter()

    def __getitem__(self, idx):
        """Syntactic sugar for accessing the counts directly."""
        return self.counts[idx]

    def update(self, y_true, y_pred, sample_weight=1.):
        self.counts[y_true].update({y_pred: sample_weight})
        self.class_counts.update([y_true, y_pred])
        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        self.counts[y_true].subtract({y_pred: sample_weight})
        self.class_counts.subtract([y_true, y_pred])
        return self

    @property
    def classes(self):
        return list(self.class_counts)

    def __repr__(self):

        # The classes are sorted alphabetically for reproducibility reasons
        classes = sorted(self.classes)

        if not classes:
            return ''

        # Determine the required width of each column in the table
        largest_label_len = max(len(str(c)) for c in classes)
        largest_number_len = len(str(max(max(c.values()) for c in self.counts.values())))
        width = max(largest_label_len, largest_number_len) + 2

        # Make a template to print out rows one by one
        row_format = '{:>{width}}' * (len(classes) + 1)

        # Write down the header
        table = row_format.format('', *map(str, classes), width=width) + '\n'

        # Write down the true labels row by row
        table += '\n'.join((
            row_format.format(
                str(y_true),
                *[f'{self.counts[y_true][y_pred]:{self._fmt}}' for y_pred in classes],
                width=width
            )
            for y_true in sorted(self.counts)
        ))

        return table



import abc
import math

from scipy import special

from creme import base


class BaseNB(base.MultiClassifier, abc.ABC):
    """Base Naive Bayes class."""

    @abc.abstractmethod
    def joint_log_likelihood(self, x: dict) -> float:
        """Compute the unnormalized posterior log-likelihood of x.

        The log-likelihood is `log P(c) + log P(x|c)`.

        """

    def predict_proba_one(self, x):
        """Return probabilities using the log-likelihoods."""
        jll = self.joint_log_likelihood(x)
        if not jll:
            return {}
        lse = special.logsumexp(list(jll.values()))
        return {label: math.exp(ll - lse) for label, ll in jll.items()}



from creme import stream

from . import base


class ImageSegments(base.FileDataset):
    """Image segments classification.

    This dataset contains features that describe image segments into 7 classes: brickface, sky,
    foliage, cement, window, path, and grass.

    References:
        1. [UCI page](https://archive.ics.uci.edu/ml/datasets/Statlog+(Image+Segmentation))

    """

    def __init__(self):
        super().__init__(
            n_samples=2310,
            n_features=18,
            task=base.MULTI_CLF,
            filename='segment.csv.zip'
        )

    def __iter__(self):
        return stream.iter_csv(
            self.path,
            target='category',
            converters={
                'region-centroid-col': int,
                'region-centroid-row': int,
                'short-line-density-5': float,
                'short-line-density-2': float,
                'vedge-mean': float,
                'vegde-sd': float,
                'hedge-mean': float,
                'hedge-sd': float,
                'intensity-mean': float,
                'rawred-mean': float,
                'rawblue-mean': float,
                'rawgreen-mean': float,
                'exred-mean': float,
                'exblue-mean': float,
                'exgreen-mean': float,
                'value-mean': float,
                'saturation-mean': float,
                'hue-mean': float
            }
        )



import collections
import copy
import functools

from .. import base
from .. import optim
from .. import utils


__all__ = ['SoftmaxRegression']


class SoftmaxRegression(base.MultiClassifier):
    """Softmax regression is a generalization of logistic regression to multiple classes.

    Softmax regression is also known as "multinomial logistic regression". There are a set weights
    for each class, hence the `weights` attribute is a nested `collections.defaultdict`. The main
    advantage of using this instead of a one-vs-all logistic regression is that the probabilities
    will be calibrated. Moreover softmax regression is more robust to outliers.

    Parameters:
        optimizer: The sequential optimizer used to tune the weights.
        loss: The loss function to optimize for.
        l2: Amount of L2 regularization used to push weights towards 0.

    Attributes:
        weights (collections.defaultdict)

    Example:

        >>> from creme import datasets
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import optim
        >>> from creme import preprocessing

        >>> X_y = datasets.ImageSegments()

        >>> model = preprocessing.StandardScaler()
        >>> model |= linear_model.SoftmaxRegression()

        >>> metric = metrics.MacroF1()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        MacroF1: 0.818765

    References:
        1. [Course on classification stochastic gradient descent](https://www.inf.ed.ac.uk/teaching/courses/mlp/2016/mlp02-sln.pdf)
        2. [Binary vs. Multi-Class Logistic Regression](https://chrisyeh96.github.io/2018/06/11/logistic-regression.html)

    """

    def __init__(self, optimizer: optim.Optimizer = None, loss: optim.losses.MultiClassLoss = None,
                 l2=0):
        if optimizer is None:
            optimizer = optim.SGD(0.01)
        new_optimizer = functools.partial(copy.deepcopy, optimizer)
        self.optimizers = collections.defaultdict(new_optimizer)  # type: ignore
        self.loss = optim.losses.CrossEntropy() if loss is None else loss
        self.l2 = l2
        self.weights = collections.defaultdict(functools.partial(collections.defaultdict, float))  # type: ignore

    def fit_one(self, x, y):

        # Some optimizers need to do something before a prediction is made
        for label, weights in self.weights.items():
            self.optimizers[label].update_before_pred(w=weights)

        # Make a prediction for the given features
        y_pred = self.predict_proba_one(x)

        # Compute the gradient of the loss w.r.t. each label
        loss_gradients = self.loss.gradient(y_true=y, y_pred=y_pred)

        for label, loss in loss_gradients.items():

            # Compute the gradient w.r.t. each feature
            weights = self.weights[label]
            gradient = {i: xi * loss + self.l2 * weights.get(i, 0) for i, xi in x.items()}
            self.weights[label] = self.optimizers[label].update_after_pred(w=weights, g=gradient)

        return self

    def predict_proba_one(self, x):
        return utils.math.softmax({
            label: utils.math.dot(weights, x)
            for label, weights in self.weights.items()
        })



"""Model selection."""
from .grid import expand_param_grid
from .score import progressive_val_score
from .sh import SuccessiveHalvingClassifier
from .sh import SuccessiveHalvingRegressor


__all__ = [
    'expand_param_grid',
    'progressive_val_score',
    'SuccessiveHalvingClassifier',
    'SuccessiveHalvingRegressor'
]



"""Stochastic optimization."""
from . import initializers
from . import losses  # type: ignore
from . import schedulers
from .ada_bound import AdaBound
from .ada_delta import AdaDelta
from .ada_grad import AdaGrad
from .ada_max import AdaMax
from .adam import Adam
from .ams_grad import AMSGrad
from .base import Optimizer
from .ftrl import FTRLProximal
from .mini_batch import MiniBatcher
from .momentum import Momentum
from .nesterov import NesterovMomentum
from .nadam import Nadam
from .rms_prop import RMSProp
from .sgd import SGD


__all__ = [
    'AdaBound',
    'AdaDelta',
    'AdaGrad',
    'Adam',
    'AMSGrad',
    'AdaMax',
    'FTRLProximal',
    'initializers',
    'losses',
    'MiniBatcher',
    'Momentum',
    'Nadam',
    'NesterovMomentum',
    'Optimizer',
    'RMSProp',
    'schedulers',
    'SGD'
]



import sklearn.utils

from creme import base
from creme import stream


def iter_sklearn_dataset(dataset: 'sklearn.utils.Bunch', **kwargs) -> base.typing.Stream:
    """Iterates rows from one of the datasets provided by scikit-learn.

    This allows you to use any dataset from [scikit-learn's `datasets` module](https://scikit-learn.org/stable/modules/classes.html#module-sklearn.datasets). For instance, you can use the `fetch_openml` function to get access to all of the
    datasets from the OpenML website.

    Parameters:
        dataset: A scikit-learn dataset.

    Example:

        >>> import pprint
        >>> from sklearn import datasets
        >>> from creme import stream

        >>> dataset = datasets.load_boston()

        >>> for xi, yi in stream.iter_sklearn_dataset(dataset):
        ...     pprint.pprint(xi)
        ...     print(yi)
        ...     break
        {'AGE': 65.2,
         'B': 396.9,
         'CHAS': 0.0,
         'CRIM': 0.00632,
         'DIS': 4.09,
         'INDUS': 2.31,
         'LSTAT': 4.98,
         'NOX': 0.538,
         'PTRATIO': 15.3,
         'RAD': 1.0,
         'RM': 6.575,
         'TAX': 296.0,
         'ZN': 18.0}
        24.0

    """
    kwargs['X'] = dataset.data
    kwargs['y'] = dataset.target
    try:
        kwargs['feature_names'] = dataset.feature_names
    except AttributeError:
        pass

    yield from stream.iter_array(**kwargs)



from .. import optim

from . import base


__all__ = ['CrossEntropy']


class CrossEntropy(base.MeanMetric, base.MultiClassMetric):
    """Multiclass generalization of the logarithmic loss.

    Example:

        >>> from creme import metrics

        >>> y_true = [0, 1, 2, 2]
        >>> y_pred = [
        ...     {0: 0.29450637, 1: 0.34216758, 2: 0.36332605},
        ...     {0: 0.21290077, 1: 0.32728332, 2: 0.45981591},
        ...     {0: 0.42860913, 1: 0.33380113, 2: 0.23758974},
        ...     {0: 0.44941979, 1: 0.32962558, 2: 0.22095463}
        ... ]

        >>> metric = metrics.CrossEntropy()

        >>> for y_t, y_p in zip(y_true, y_pred):
        ...     metric = metric.update(y_t, y_p)
        ...     print(metric.get())
        1.222454
        1.169691
        1.258864
        1.321597

        >>> metric
        CrossEntropy: 1.321598

    """

    @property
    def bigger_is_better(self):
        return False

    @property
    def requires_labels(self):
        return False

    def _eval(self, y_true, y_pred):
        return optim.losses.CrossEntropy()(y_true, y_pred)



"""Learning rate schedulers."""
import abc
import math

from creme import optim


__all__ = [
    'Constant',
    'InverseScaling',
    'Optimal',
    'Scheduler'
]


class Scheduler(abc.ABC):

    @abc.abstractmethod
    def get(self, t: int) -> float:
        """Returns the learning rate at a given iteration.

        Parameters:
            t: The iteration number.

        """

    def __str__(self):
        return self.__class__.__name__

    def __repr__(self):
        return f'{self.__class__.__name__}({vars(self)})'


class Constant(Scheduler):
    """Always uses the same learning rate.

    Parameters:
        learning_rate

    """

    def __init__(self, learning_rate: float):
        self.learning_rate = learning_rate

    def get(self, t):
        return self.learning_rate


class InverseScaling(Scheduler):
    """Reduces the learning rate using a power schedule.

    Assuming an initial learning rate $\eta$, the learning rate at step $t$ is:

    $$\\frac{eta}{(t + 1) ^ p}$$

    where $p$ is a user-defined parameter.

    Parameters:
        learning_rate
        power

    """

    def __init__(self, learning_rate: float, power=0.5):
        self.learning_rate = learning_rate
        self.power = power

    def get(self, t):
        return self.learning_rate / pow(t + 1, self.power)


class Optimal(Scheduler):
    """Optimal learning schedule as proposed by Léon Bottou.

    Parameters:
        loss
        alpha

    References:
        1. [Bottou, L., 2012. Stochastic gradient descent tricks. In Neural networks: Tricks of the trade (pp. 421-436). Springer, Berlin, Heidelberg.](https://cilvr.cs.nyu.edu/diglib/lsml/bottou-sgd-tricks-2012.pdf)

    """

    def __init__(self, loss: optim.losses.Loss, alpha=1e-4):
        self.loss = loss
        self.alpha = alpha

        typw = math.sqrt(1. / math.sqrt(self.alpha))
        initial_eta0 = typw / max(1.0, self.loss.gradient(-typw, 1.0))
        self.t0 = 1. / (initial_eta0 * self.alpha)

    def get(self, t):
        return 1. / (self.alpha * (self.t0 + t))



"""Time series forecasting."""
from .detrender import Detrender
from .detrender import GroupDetrender
from .snarimax import SNARIMAX


__all__ = [
    'Detrender',
    'GroupDetrender',
    'SNARIMAX'
]



import math
import random

import mmh3

from . import base


class NUnique(base.Univariate):
    """Approximate number of unique values counter.

    This is basically an implementation of the HyperLogLog algorithm. Adapted from
    [`hypy`](https://github.com/clarkduvall/hypy). The code is a bit too terse but it will do for
    now.

    Parameters:
        error_rate: Desired error rate. Memory usage is inversely proportional to this value.
        seed: Set the seed to produce identical results.

    Attributes:
        n_bits (int)
        n_buckets (int)
        buckets (list)

    Example:

        >>> import string
        >>> from creme import stats

        >>> alphabet = string.ascii_lowercase
        >>> n_unique = stats.NUnique(error_rate=0.1, seed=42)

        >>> n_unique.update('a').get()
        1

        >>> n_unique.update('b').get()
        2

        >>> for letter in alphabet:
        ...     n_unique = n_unique.update(letter)
        >>> n_unique.get()
        27

        We can increase the precision by lowering the ``error_rate`` parameter.

        >>> n_unique = stats.NUnique(error_rate=0.01, seed=42)
        >>> for letter in alphabet:
        ...     n_unique = n_unique.update(letter)
        >>> n_unique.get()
        26

    References:
        1. [My favorite algorithm (and data structure): HyperLogLog](https://odino.org/my-favorite-data-structure-hyperloglog/)
        2. [Flajolet, P., Fusy, É., Gandouet, O. and Meunier, F., 2007, June. Hyperloglog: the analysis of a near-optimal cardinality estimation algorithm.](http://algo.inria.fr/flajolet/Publications/FlFuGaMe07.pdf)

    """

    P32 = 2 ** 32

    def __init__(self, error_rate=.01, seed=None):
        self.n_bits = int(math.ceil(math.log((1.04 / error_rate) ** 2, 2)))
        self.n_buckets = 1 << self.n_bits
        self.buckets = [0] * self.n_buckets
        self.seed = seed
        self._hash_seed = random.Random(seed).randint(0, 2 ** 10)

    @property
    def name(self):
        return 'n_unique'

    def update(self, x):
        x = mmh3.hash(x, seed=self._hash_seed)
        i = x & NUnique.P32 - 1 >> 32 - self.n_bits
        z = 35 - len(bin(NUnique.P32 - 1 & x << self.n_bits | 1 << self.n_bits - 1))
        self.buckets[i] = max(self.buckets[i], z)
        return self

    def get(self):
        a = ({16: 0.673, 32: 0.697, 64: 0.709}[self.m]
             if self.n_buckets <= 64 else 0.7213 / (1 + 1.079 / self.n_buckets))
        e = a * self.n_buckets * self.n_buckets / sum(1.0 / (1 << x) for x in self.buckets)
        if e <= self.n_buckets * 2.5:
            z = len([r for r in self.buckets if not r])
            return int(self.n_buckets * math.log(float(self.n_buckets) / z) if z else e)
        return int(e if e < NUnique.P32 / 30 else -NUnique.P32 * math.log(1 - e / NUnique.P32))



import itertools
import random
import typing

import numpy as np

from creme import base


def iter_array(X: np.ndarray, y: np.ndarray = None,
               feature_names: typing.List[base.typing.FeatureName] = None,
               target_names: typing.List[base.typing.FeatureName] = None, shuffle: bool = False,
               seed: int = None) -> base.typing.Stream:
    """Iterates over the rows from an array of features and an array of targets.

    This method is intended to work with `numpy` arrays, but should also work with Python lists.

    Parameters:
        X: A 2D array of features.
        y: An optional array of targets.
        feature_names: An optional list of feature names. The features will be labeled with
            integers if no names are provided.
        target_names: An optional list of output names. The outputs will be labeled with integers
            if no names are provided. Only applies if there are multiple outputs, i.e. if `y` is a
            2D array.
        shuffle: Indicates whether or not to shuffle the input arrays before iterating over them.
        seed: Random seed used for shuffling the data.

    Example:

        >>> from creme import stream
        >>> import numpy as np

        >>> X = np.array([[1, 2, 3], [11, 12, 13]])
        >>> Y = np.array([True, False])

        >>> X_y = stream.iter_array(
        ...     X, Y,
        ...     feature_names=['x1', 'x2', 'x3']
        ... )
        >>> for x, y in X_y:
        ...     print(x, y)
        {'x1': 1, 'x2': 2, 'x3': 3} True
        {'x1': 11, 'x2': 12, 'x3': 13} False

    """
    feature_names = list(range(len(X[0]))) if feature_names is None else feature_names

    multioutput = y is not None and not np.isscalar(y[0])
    if multioutput and target_names is None:
        target_names = list(range(len(y[0])))

    # Shuffle the data
    rng = random.Random(seed)
    if shuffle:
        order = rng.sample(range(len(X)), k=len(X))
        X = X[order]
        y = y if y is None else y[order]

    if multioutput:

        for xi, yi in itertools.zip_longest(X, y if hasattr(y, '__iter__') else []):
            yield dict(zip(feature_names, xi)), dict(zip(target_names, yi))

    else:

        for xi, yi in itertools.zip_longest(X, y if hasattr(y, '__iter__') else []):
            yield dict(zip(feature_names, xi)), yi



import collections
import typing
import random

from creme import base
from creme import optim
from creme import utils


class Triplet(collections.namedtuple('Triplet', 'x y loss')):

    def __lt__(self, other):
        return self.loss < other.loss


class HardSampling(base.Wrapper):
    """Hard sampler."""

    def __init__(self, model, loss, size, p,seed = None):
        self.model = model
        self.loss = loss
        self.pred_func = model.predict_one

        if isinstance(model, base.Classifier):
            self.pred_func = model.predict_proba_one

        elif isinstance(model, base.BinaryClassifier):
            self.pred_func = lambda x: model.predict_proba_one(x)[True]

        self.p = p
        self.size = size
        self.buffer = utils.SortedWindow(self.size)
        self.seed = seed
        self._rng = random.Random(seed)

    @property
    def _wrapped_model(self):
        return self.model

    def predict_one(self, x):
        return self.model.predict_one(x)

    def fit_one(self, x, y):

        loss = self.loss(y_true=y, y_pred=self.pred_func(x))

        if len(self.buffer) < self.size:
            self.buffer.append(Triplet(x=x, y=y, loss=loss))

        elif loss > self.buffer[0].loss:
            self.buffer.pop(0)

            self.buffer.append(Triplet(x=x, y=y, loss=loss))

        # Probability p
        if self._rng.uniform(0, 1) <= self.p:
            i = self._rng.randint(0, len(self.buffer) - 1)

            triplet = self.buffer.pop(i)

            self.model.fit_one(triplet.x, triplet.y)

            loss = self.loss(y_true=triplet.y, y_pred=self.pred_func(triplet.x))

            self.buffer.append(Triplet(x=triplet.x, y=triplet.y, loss=loss))

        # Probability (1 - p)
        else:
            self.model.fit_one(x, y)

        return self


class HardSamplingRegressor(HardSampling):
    """Hard sampling regressor.

    This wrapper enables a model to retrain on past samples who's output was hard to predict.
    This works by storing the hardest samples in a buffer of a fixed size. When a new sample
    arrives, the wrapped model is either trained on one of the buffered samples with a probability
    p or on the new sample with a probability (1 - p).

    The hardness of an observation is evaluated with a loss function that compares the sample's
    ground truth with the wrapped model's prediction. If the buffer is not full, then the sample
    is added to the buffer. If the buffer is full and the new sample has a bigger loss than the
    lowest loss in the buffer, then the sample takes it's place.

    Parameters:
        regressor
        size: Size of the buffer.
        p: Probability of updating the model with a sample from the buffer instead of a new
            incoming sample.
        loss: Criterion used to evaluate the hardness of a sample.
        seed: Random seed.

    Example:

        >>> from creme import datasets
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import optim
        >>> from creme import preprocessing
        >>> from creme import sampling

        >>> model = preprocessing.StandardScaler()

        >>> model = (
        ...     preprocessing.StandardScaler() |
        ...     sampling.HardSamplingRegressor(
        ...         regressor=linear_model.LinearRegression(),
        ...         p=.2,
        ...         size=30,
        ...         seed=42,
        ...     )
        ... )

        >>> model_selection.progressive_val_score(
        ...     datasets.TrumpApproval(),
        ...     model,
        ...     metrics.MAE(),
        ...     print_every=500
        ... )
        [500] MAE: 2.292501
        [1,000] MAE: 1.395797
        MAE: 1.394693

    """

    def __init__(self, regressor: base.Regressor, size: int, p: float,
                 loss: optim.losses.RegressionLoss = None, seed: int = None):
        if loss is None:
            loss =  optim.losses.Absolute()
        super().__init__(model=regressor, loss=loss, size=size, p=p, seed=seed)


class HardSamplingClassifier(HardSampling):
    """Hard sampling classifier.

    This wrapper enables a model to retrain on past samples who's output was hard to predict.
    This works by storing the hardest samples in a buffer of a fixed size. When a new sample
    arrives, the wrapped model is either trained on one of the buffered samples with a probability
    p or on the new sample with a probability (1 - p).

    The hardness of an observation is evaluated with a loss function that compares the sample's
    ground truth with the wrapped model's prediction. If the buffer is not full, then the sample
    is added to the buffer. If the buffer is full and the new sample has a bigger loss than the
    lowest loss in the buffer, then the sample takes it's place.

    Parameters:
        classifier
        size: Size of the buffer.
        p: Probability of updating the model with a sample from the buffer instead of a new
            incoming sample.
        loss: Criterion used to evaluate the hardness of a sample.
        seed: Random seed.

    Example:

        >>> from creme import datasets
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import optim
        >>> from creme import preprocessing
        >>> from creme import sampling

        >>> model = preprocessing.StandardScaler()

        >>> model = (
        ...     preprocessing.StandardScaler() |
        ...     sampling.HardSamplingClassifier(
        ...         classifier=linear_model.LogisticRegression(),
        ...         p=0.1,
        ...         size=40,
        ...         seed=42,
        ...     )
        ... )

        >>> model_selection.progressive_val_score(
        ...     X_y=datasets.Phishing(),
        ...     model=model,
        ...     metric=metrics.ROCAUC(),
        ...     print_every=500,
        ... )
        [500] ROCAUC: 0.928476
        [1,000] ROCAUC: 0.948367
        ROCAUC: 0.952511

    """

    def __init__(self, classifier: base.Classifier, size: int, p: float,
                 loss: typing.Union[optim.losses.BinaryLoss, optim.losses.MultiClassLoss] = None,
                 seed: int = None):
        if loss is None:
            loss = optim.losses.CrossEntropy()
        super().__init__(model=classifier, loss=loss, size=size, p=p, seed=seed)

    def predict_proba_one(self, x):
        return self.model.predict_proba_one(x)



"""Compatibility tools.

This module contains adapters for making ``creme`` estimators compatible with other libraries, and
vice-versa whenever possible. The relevant adapters will only be usable if you have installed the
necessary library. For instance, to use the ``compat.convert_sklearn_to_creme`` function, you have
to install scikit-learn.

"""
import typing

__all__: typing.List[str] = []

try:
    from .sklearn import convert_creme_to_sklearn
    from .sklearn import convert_sklearn_to_creme
    from .sklearn import SKL2CremeClassifier
    from .sklearn import SKL2CremeRegressor
    from .sklearn import Creme2SKLRegressor
    from .sklearn import Creme2SKLClassifier
    from .sklearn import Creme2SKLClusterer
    from .sklearn import Creme2SKLTransformer

    __all__ += [
        'convert_creme_to_sklearn',
        'convert_sklearn_to_creme',
        'Creme2SKLRegressor',
        'Creme2SKLClassifier',
        'Creme2SKLClusterer',
        'Creme2SKLTransformer',
        'SKL2CremeClassifier',
        'SKL2CremeRegressor'
    ]
except ModuleNotFoundError:
    pass

try:
    from .pytorch import PyTorch2CremeRegressor

    __all__ += ['PyTorch2CremeRegressor']
except ModuleNotFoundError:
    pass




from .. import utils

from . import base
from . import report


__all__ = ['Rolling']


class Rolling(base.WrapperMetric, utils.Window):
    """Wrapper for computing metrics over a window.

    This wrapper metric allows you to apply a metric over a window of observations. Under the hood,
    a buffer with the `window_size` most recent pairs of `(y_true, y_pred)` is memorised. When the
    buffer is full, the oldest pair is removed and the `revert` method of the metric is called with
    said pair.

    Parameters:
        metric: A metric.
        window_size: The number of most recent `(y_true, y_pred)` pairs on which to evaluate the
            metric.

    Example:

        >>> from creme import metrics

        >>> y_true = [3, -0.5, 2, 7]
        >>> y_pred = [2.5, 0.0, 2, 8]

        >>> metric = metrics.Rolling(metrics.MSE(), window_size=2)

        >>> for yt, yp in zip(y_true, y_pred):
        ...     print(metric.update(yt, yp))
        Rolling of size 2 MSE: 0.25
        Rolling of size 2 MSE: 0.25
        Rolling of size 2 MSE: 0.125
        Rolling of size 2 MSE: 0.5

    .. tip::
        You should use `Rolling` to evaluate a metric over a window of fixed sized. You can use
        `creme.metrics.TimeRolling` to instead evaluate a metric over a period of time.

    """

    def __init__(self, metric: base.Metric, window_size: int):
        super().__init__(size=window_size)
        self.window_size = window_size
        self._metric = metric

    @property
    def metric(self):
        return self._metric

    def update(self, y_true, y_pred, sample_weight=1.):
        self.metric.update(y_true, y_pred, sample_weight)
        if len(self) == self.window_size:
            self.metric.revert(*self[0])
        self.append((y_true, y_pred, sample_weight))
        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        self.metric.revert(y_true, y_pred, sample_weight)
        return self

    def __repr__(self):
        if isinstance(self.metric, report.ClassificationReport):
            return self.metric.__repr__()
        return f'Rolling of size {self.window_size} {str(self.metric)}'



import math
import typing

from creme import base
from creme import optim


__all__ = ['HedgeRegressor']


class HedgeRegressor(base.Ensemble, base.Regressor):
    """Hedge Algorithm for regression.

    The Hedge Algorithm is a special case of the Weighted Majority Algorithm for arbitrary losses.

    Parameters:
        regressors: The regressors to hedge.
        loss: The loss function that has to be minimized. Defaults to `optim.losses.Squared`.
        learning_rate: The learning rate by which the model weights are multiplied at each
            iteration.

    Example:

        >>> from creme import datasets
        >>> from creme import ensemble
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import optim
        >>> from creme import preprocessing

        >>> optimizers = [
        ...     optim.SGD(0.01),
        ...     optim.RMSProp(),
        ...     optim.AdaGrad()
        ... ]

        >>> for optimizer in optimizers:
        ...
        ...     X_y = datasets.TrumpApproval()
        ...     metric = metrics.MAE()
        ...     model = (
        ...         preprocessing.StandardScaler() |
        ...         linear_model.LinearRegression(
        ...             optimizer=optimizer,
        ...             intercept_lr=.1
        ...         )
        ...     )
        ...
        ...     print(optimizer, model_selection.progressive_val_score(X_y, model, metric))
        SGD MAE: 0.555971
        RMSProp MAE: 0.528284
        AdaGrad MAE: 0.481461

        >>> X_y = datasets.TrumpApproval()
        >>> metric = metrics.MAE()
        >>> hedge = (
        ...     preprocessing.StandardScaler() |
        ...     ensemble.HedgeRegressor(
        ...         regressors=[
        ...             linear_model.LinearRegression(optimizer=o, intercept_lr=.1)
        ...             for o in optimizers
        ...         ],
        ...         learning_rate=0.005
        ...     )
        ... )

        >>> model_selection.progressive_val_score(X_y, hedge, metric)
        MAE: 0.494832

    References:
        1. [Online Learning from Experts: Weighed Majority and Hedge](https://www.shivani-agarwal.net/Teaching/E0370/Aug-2011/Lectures/20-scribe1.pdf)
        2. [Wikipedia page on the multiplicative weight update method](https://www.wikiwand.com/en/Multiplicative_weight_update_method)
        3. [Kivinen, J. and Warmuth, M.K., 1997. Exponentiated gradient versus gradient descent for linear predictors. information and computation, 132(1), pp.1-63.](https://users.soe.ucsc.edu/~manfred/pubs/J36.pdf)

    """

    def __init__(self, regressors: typing.List[base.Regressor],
                 loss: optim.losses.RegressionLoss = None, learning_rate=.5):
        super().__init__(regressors)
        self.loss = optim.losses.Squared() if loss is None else loss
        self.learning_rate = learning_rate
        self.weights = [1.] * len(regressors)

    @property
    def regressors(self):
        return self.models

    def fit_predict_one(self, x, y):

        y_pred_mean = 0.

        # Make a prediction and update the weights accordingly for each model
        total = 0
        for i, regressor in enumerate(self):
            y_pred = regressor.predict_one(x=x)
            y_pred_mean += self.weights[i] * (y_pred - y_pred_mean) / len(self)
            loss = self.loss(y_true=y, y_pred=y_pred)
            self.weights[i] *= math.exp(-self.learning_rate * loss)
            total += self.weights[i]
            regressor.fit_one(x, y)

        # Normalize the weights so that they sum up to 1
        if total:
            for i, _ in enumerate(self.weights):
                self.weights[i] /= total

        return y_pred_mean

    def fit_one(self, x, y):
        self.fit_predict_one(x, y)
        return self

    def predict_one(self, x):
        return sum(model.predict_one(x) * weight for model, weight in zip(self, self.weights))



import pytest
from sklearn.utils import estimator_checks

from creme import base
from creme import cluster
from creme import compat
from creme import linear_model
from creme import preprocessing
from sklearn import linear_model as sk_linear_model


@pytest.mark.parametrize('estimator', [
    pytest.param(estimator, id=str(estimator))
    for estimator in [
        linear_model.LinearRegression(),
        preprocessing.StandardScaler(),
        cluster.KMeans(seed=42)
    ]
])
def test_sklearn_check_estimator(estimator: base.Estimator):
    skl_estimator = compat.convert_creme_to_sklearn(estimator)
    estimator_checks.check_estimator(skl_estimator)



"""Base interfaces.

Every estimator in `creme` is a class, and as such inherits from at least one base interface.
These are used to categorize, organize, and standardize the many estimators that `creme`
contains.

"""
from . import typing
from .anomaly import AnomalyDetector
from .classifier import Classifier
from .classifier import BinaryClassifier
from .classifier import MultiClassifier
from .clusterer import Clusterer
from .ensemble import Ensemble
from .estimator import Estimator
from .multi_output import MultiOutputClassifier
from .multi_output import MultiOutputRegressor
from .predictor import Predictor
from .regressor import Regressor
from .transformer import SupervisedTransformer
from .transformer import Transformer
from .wrapper import Wrapper


__all__ = [
    'AnomalyDetector',
    'BinaryClassifier',
    'Classifier',
    'Clusterer',
    'Ensemble',
    'Estimator',
    'MultiClassifier',
    'MultiOutputClassifier',
    'MultiOutputRegressor',
    'Predictor',
    'Regressor',
    'SupervisedTransformer',
    'Transformer',
    'typing',
    'Wrapper'
]
